@@ -10,6 +10,12 @@
 /* Have strerror() function. */
 #undef HAVE_STRERROR
 
+/* Have memmove() function. */
+#undef HAVE_MEMMOVE
+
+/* Have memcpy() function. */
+#undef HAVE_MEMCPY
+
 /* How to retrieve the current working directory? */
 #undef HAVE_GETCWD
 #undef HAVE_GETWD
