@@ -49,9 +49,9 @@ static char *error_names[] =
  * Prototypes for static functions.
  */
 
-static void read_font_map __P ((AFMHandle handle, char *name));
-static void apply_encoding __P ((AFMFont font, AFMEncodingTable *enc,
-				 unsigned int flags));
+static void read_font_map ___P ((AFMHandle handle, char *name));
+static void apply_encoding ___P ((AFMFont font, AFMEncodingTable *enc,
+				  unsigned int flags));
 
 
 /*
@@ -630,6 +630,10 @@ afm_font_encoding (AFMFont font, AFMEncoding enc, unsigned int flags)
       apply_encoding (font, afm_latin3_encoding, flags);
       break;
 
+    case AFM_ENCODING_LATIN5:
+      apply_encoding (font, afm_latin5_encoding, flags);
+      break;
+
     case AFM_ENCODING_IBMPC:
       apply_encoding (font, afm_ibmpc_encoding, flags);
       break;
@@ -655,6 +659,10 @@ afm_font_encoding (AFMFont font, AFMEncoding enc, unsigned int flags)
     case AFM_ENCODING_HP8:
       apply_encoding (font, afm_hp8_encoding, flags);
       break;
+
+    case AFM_ENCODING_KOI8:
+      apply_encoding (font, afm_koi8_encoding, flags);
+      break;
     }
 
   return AFM_SUCCESS;
