@@ -151,22 +151,22 @@ static struct keyname_st
  */
 
 /* Throw parse error <error>.  Never returns. */
-static void parse_error __P ((AFMHandle handle, AFMError error));
+static void parse_error ___P ((AFMHandle handle, AFMError error));
 
-static int get_token __P ((AFMHandle handle, ParseCtx *ctx));
-static int get_line_token __P ((AFMHandle handle, ParseCtx *ctx));
-static void get_key __P ((AFMHandle handle, ParseCtx *ctx,
-			  AFMKey *key_return));
-static void get_type __P ((AFMHandle handle, ParseCtx *ctx, int type,
-			   AFMNode *type_return));
-static void read_character_metrics __P ((AFMHandle handle, ParseCtx *ctx,
-					 AFMFont font));
-static void read_kern_pairs __P ((AFMHandle handle, ParseCtx *ctx,
-				  AFMFont font));
-static void read_track_kerns __P ((AFMHandle handle, ParseCtx *ctx,
+static int get_token ___P ((AFMHandle handle, ParseCtx *ctx));
+static int get_line_token ___P ((AFMHandle handle, ParseCtx *ctx));
+static void get_key ___P ((AFMHandle handle, ParseCtx *ctx,
+			   AFMKey *key_return));
+static void get_type ___P ((AFMHandle handle, ParseCtx *ctx, int type,
+			    AFMNode *type_return));
+static void read_character_metrics ___P ((AFMHandle handle, ParseCtx *ctx,
+					  AFMFont font));
+static void read_kern_pairs ___P ((AFMHandle handle, ParseCtx *ctx,
+				   AFMFont font));
+static void read_track_kerns ___P ((AFMHandle handle, ParseCtx *ctx,
+				    AFMFont font));
+static void read_composites ___P ((AFMHandle handle, ParseCtx *ctx,
 				   AFMFont font));
-static void read_composites __P ((AFMHandle handle, ParseCtx *ctx,
-				  AFMFont font));
 
 /*
  * Global functions.
