@@ -37,11 +37,14 @@
 
 #include <stdio.h>
 
-#ifndef __P
+#include <sys/types.h>
+#include <sys/stat.h>
+
+#ifndef ___P
 #if PROTOTYPES
-#define __P(protos) protos
+#define ___P(protos) protos
 #else /* no PROTOTYPES */
-#define __P(protos) ()
+#define ___P(protos) ()
 #endif /* no PROTOTYPES */
 #endif
 
@@ -68,11 +71,6 @@
 char *strchr ();
 char *strrchr ();
 
-#ifndef HAVE_MEMCPY
-#define memcpy(d, s, n) bcopy((s), (d), (n))
-#define memmove(d, s, n) bcopy((s), (d), (n))
-#endif
-
 #if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
 #include <stdarg.h>
 #else
@@ -80,7 +78,15 @@ char *strrchr ();
 #endif
 
 #ifndef HAVE_STRERROR
-extern char *strerror __P ((int));
+extern char *strerror ___P ((int));
+#endif
+
+#ifndef HAVE_MEMMOVE
+extern void *memmove ___P ((void *, void *, size_t));
+#endif
+
+#ifndef HAVE_MEMCPY
+extern void *memcpy ___P ((void *, void *, size_t));
 #endif
 
 #endif /* no STDC_HEADERS */
@@ -92,15 +98,13 @@ extern char *strerror __P ((int));
 #if HAVE_MATH_H
 #include <math.h>
 #else
-extern double atan2 __P ((double, double));
+extern double atan2 ___P ((double, double));
 #endif
 
 #include <errno.h>
 #include <time.h>
 #include <assert.h>
-#include <sys/types.h>
 #include <ctype.h>
-#include <sys/stat.h>
 
 #if HAVE_PWD_H
 #include <pwd.h>
@@ -127,6 +131,7 @@ extern double atan2 __P ((double, double));
 
 #include "afm.h"
 #include "strhash.h"
+#include "xalloc.h"
 
 /* 
  * Types and definitions.
@@ -141,7 +146,7 @@ extern double atan2 __P ((double, double));
 #define CHAR_WIDTH(ch) (font_widths[(unsigned char) (ch)])
 
 /* Current point y movement from line to line. */
-#define LINESKIP (Fpt + baselineskip)
+#define LINESKIP (Fpt.h + baselineskip)
 
 
 /* Constants for output files. */
@@ -179,12 +184,15 @@ typedef enum
   ENC_LATIN1,
   ENC_LATIN2,
   ENC_LATIN3,
+  ENC_LATIN5,
   ENC_ASCII,
-  ENC_ASCII_SCANDS,
+  ENC_ASCII_FISE,
+  ENC_ASCII_DKNO,
   ENC_IBMPC,
   ENC_MAC,
   ENC_VMS,
   ENC_HP8,
+  ENC_KOI8,
   ENC_PS
 } InputEncoding;
 
@@ -217,7 +225,7 @@ struct file_lookup_ctx_st
 
 typedef struct file_lookup_ctx_st FileLookupCtx;
 
-typedef int (*PathWalkProc) __P ((char *path, void *context));
+typedef int (*PathWalkProc) ___P ((char *path, void *context));
 
 struct input_stream_st
 {
@@ -243,6 +251,33 @@ struct page_range_st
 
 typedef struct page_range_st PageRange;
 
+struct font_point_st
+{
+  double w;			/* width */
+  double h;			/* height */
+};
+
+typedef struct font_point_st FontPoint;
+
+struct color_st
+{
+  float r;
+  float g;
+  float b;
+};
+
+typedef struct color_st Color;
+
+struct cached_font_info_st
+{
+  double font_widths[256];
+  char font_ctype[256];
+  AFMBoolean font_is_fixed;
+  AFMNumber font_bbox_lly;
+};
+
+typedef struct cached_font_info_st CachedFontInfo;
+
 
 /* 
  * Global variables.
@@ -261,11 +296,16 @@ extern char *afm_path;
 extern char afm_path_buffer[];
 extern MediaEntry *media_names;
 extern MediaEntry *media;
-extern char spooler_command[];
+extern char no_job_header_switch[];
+extern char output_first_line[];
 extern char queue_param[];
+extern char spooler_command[];
 extern int nl;
 extern int bs;
 extern unsigned int current_pagenum;
+extern unsigned int input_filenum;
+extern unsigned int current_file_linenum;
+extern char fname[];
 
 /* Statistics. */
 extern int total_pages;
@@ -301,16 +341,17 @@ extern StringHashPtr user_strings;
 
 /* Cache for AFM files. */
 extern StringHashPtr afm_cache;
+extern StringHashPtr afm_info_cache;
 
 /* AFM library handle. */
 extern AFMHandle afm;
 
 /* Fonts. */
 extern char *HFname;
-extern double HFpt;
+extern FontPoint HFpt;
 extern char *Fname;
-extern double Fpt;
-extern double default_Fpt;
+extern FontPoint Fpt;
+extern FontPoint default_Fpt;
 extern char *default_Fname;
 
 extern double font_widths[];
@@ -345,7 +386,7 @@ extern int special_escapes;
 extern int escape_char;
 extern int tabsize;
 extern double baselineskip;
-extern double ul_ptsize;
+extern FontPoint ul_ptsize;
 extern double ul_gray;
 extern char *ul_font;
 extern char *underlay;
@@ -379,6 +420,22 @@ extern double line_highlight_gray;
 extern int accept_composites;
 extern FormFeedType formfeed_type;
 extern char *input_filter_stdin;
+extern int toc;
+extern FILE *toc_fp;
+extern char *toc_fmt_string;
+extern unsigned int file_align;
+extern int slicing;
+extern unsigned int slice;
+extern char states_color_model[];
+extern char states_config_file[];
+extern char states_highlight_level[];
+extern unsigned int nup;
+extern unsigned int nup_rows;
+extern unsigned int nup_columns;
+extern int nup_landscape;
+extern unsigned int nup_width;
+extern unsigned int nup_height;
+extern double nup_scale;
 
 
 /* 
@@ -386,79 +443,79 @@ extern char *input_filter_stdin;
  */
 
 /* Print message if <verbose> is >= <verbose_level>. */
-void message __P ((int verbose_level, char *fmt, ...));
+void message ___P ((int verbose_level, char *fmt, ...));
 
 /* Report continuable error. */
-void error __P ((char *fmt, ...));
+void error ___P ((char *fmt, ...));
 
 /* Report fatal error and exit with status 1.  Function never returns. */
-void fatal __P ((char *fmt, ...));
+void fatal ___P ((char *fmt, ...));
 
 /* 
  * Read config file <path, name>.  Returns bool.  If function fails, error
  * is found from errno.
  */
-int read_config __P ((char *path, char *name));
+int read_config ___P ((char *path, char *name));
 
 /* Print PostScript header to our output stream. */
-void dump_ps_header __P ((void));
+void dump_ps_header ___P ((void));
 
 /* Print PostScript trailer to our output stream. */
-void dump_ps_trailer __P ((void));
+void dump_ps_trailer ___P ((void));
 
 /*
  * Open InputStream to <fp> or <fname>.  If <input_filter> is given
  * it is used to pre-filter the incoming data stream.  Function returns
  * 1 if stream could be opened or 0 otherwise.
  */
-int is_open __P ((InputStream *is, FILE *fp, char *fname,
-		  char *input_filter));
+int is_open ___P ((InputStream *is, FILE *fp, char *fname,
+		   char *input_filter));
 
 /* Close InputStream <is>. */
-void is_close __P ((InputStream *is));
+void is_close ___P ((InputStream *is));
 
 /*
  * Read next character from the InputStream <is>.  Returns EOF if 
  * EOF was reached.
  */
-int is_getc __P ((InputStream *is));
+int is_getc ___P ((InputStream *is));
 
 /*
  * Put character <ch> back to the InputStream <is>.  Function returns EOF
  * if character couldn't be unget.
  */
-int is_ungetc __P ((int ch, InputStream *is));
+int is_ungetc ___P ((int ch, InputStream *is));
 
 /* 
  * Process single input file <fp>.  File's name is given in <fname> and
  * it is used to print headers.
  */
-void process_file __P ((char *fname, InputStream *fp));
+void process_file ___P ((char *fname, InputStream *fp));
 
 /* Add a new media to the list of known media. */
-void add_media __P ((char *name, int w, int h, int llx, int lly, int urx,
-		     int ury));
+void add_media ___P ((char *name, int w, int h, int llx, int lly, int urx,
+		      int ury));
 
 /* Print a listing of missing characters. */
-void do_list_missing_characters __P ((int *array));
+void do_list_missing_characters ___P ((int *array));
 
 /* 
  * Check if file <name, suffix> exists.  Returns bool.  If function fails
  * error can be found from errno.
  */
-int file_existsp __P ((char *name, char *suffix));
+int file_existsp ___P ((char *name, char *suffix));
 
 /*
  * Paste file <name, suffix> to output stream.  Returns bool. If 
  * function fails, error can be found from errno.
  */
-int paste_file __P ((char *name, char *suffix));
+int paste_file ___P ((char *name, char *suffix));
 
 /*
  * Do tilde substitution for filename <from> and insert result to <to>.
  * Buffer <to> must be long enought to hold expanded name.
  */
-void tilde_subst __P ((char *from, char *to));
+void tilde_subst ___P ((char *from, char *to));
 
 /*
  * Parse one float dimension from string <string>.  If <units> is true,
@@ -466,7 +523,7 @@ void tilde_subst __P ((char *from, char *to));
  * <horizontal> is true, then dimension is horizontal, otherwise it
  * is vertical.
  */
-double parse_float __P ((char *string, int units, int horizontal));
+double parse_float ___P ((char *string, int units, int horizontal));
 
 /*
  * Parse font spec <spec> and return font's name and size in variables
@@ -474,40 +531,40 @@ double parse_float __P ((char *string, int units, int horizontal));
  * font spec or 0 otherwise.  Returned name <name_return> is allocated
  * with xcalloc() and must be freed by caller.
  */
-int parse_font_spec __P ((char *spec, char **name_return,
-			  double *size_return));
+int parse_font_spec ___P ((char *spec, char **name_return,
+			   FontPoint *size_return));
 
 /*
  * Read body font's character widths and character codes from AFM files.
  */
-void read_font_info __P ((void));
+void read_font_info ___P ((void));
 
 /*
  * Try to download font <name>.
  */
-void download_font __P ((char *name));
+void download_font ___P ((char *name));
 
 /*
  * Escape all PostScript string's special characters from string <string>.
  * Returns a xmalloc()ated result.
  */
-char *escape_string __P ((char *string));
+char *escape_string ___P ((char *string));
 
 /*
  * Expand user escapes from string <string>.  Returns a xmalloc()ated
  * result.
  */
-char *format_user_string __P ((char *string));
+char *format_user_string ___P ((char *context_name, char *string));
 
 /*
  * Parses key-value pair <kv> and inserts/deletes key from <set>.
  */
-void parse_key_value_pair __P ((StringHashPtr set, char *kv));
+void parse_key_value_pair ___P ((StringHashPtr set, char *kv));
 
 /*
  * Count how many non-empty items there are in the key-value set <set>.
  */
-int count_key_value_set __P ((StringHashPtr set));
+int count_key_value_set ___P ((StringHashPtr set));
 
 /*
  * Walk through path <path> and call <proc> once for each of its
@@ -516,25 +573,10 @@ int count_key_value_set __P ((StringHashPtr set));
  * return value.  In that case value is returned as the return value
  * of the pathwalk().
  */
-int pathwalk __P ((char *path, PathWalkProc proc, void *context));
+int pathwalk ___P ((char *path, PathWalkProc proc, void *context));
 
 /* Lookup file from path.  <context> must point to FileLookupCtx. */
-int file_lookup __P ((char *path, void *context));
-
-
-/* 
- * Non-failing memory allocation.
- */
-
-void *xmalloc __P ((size_t size));
-
-void *xcalloc __P ((size_t num, size_t size));
-
-void *xrealloc __P ((void *ptr, size_t size));
-
-void xfree __P ((void *ptr));
-
-char *xstrdup __P ((char *));
+int file_lookup ___P ((char *path, void *context));
 
 
 /*
@@ -549,12 +591,12 @@ char *xstrdup __P ((char *));
  * to variable <context_return> wich is passed as an argument to the
  * printer_close() function.
  */
-FILE *printer_open __P ((char *cmd, char *options, char *queue_param,
-			 char *printer_name, void **context_return));
+FILE *printer_open ___P ((char *cmd, char *options, char *queue_param,
+			  char *printer_name, void **context_return));
 
 /*
  * Flush all pending output to printer <context> and close it.
  */
-void printer_close __P ((void *context));
+void printer_close ___P ((void *context));
 
 #endif /* not GSINT_H */
