@@ -31,11 +31,20 @@
  * Prototypes for static functions.
  */
 
+/*
+ * Open output file according to user options.  Void if output file
+ * has already been opened.
+ */
+static void open_output_file ();
+
+/* Close output file. */
+static void close_output_file ();
+
 /* Handle options from environment variable <var> */
-static void handle_env_options __P ((char *var));
+static void handle_env_options ___P ((char *var));
 
 /* Handle options from <argv> array. */
-static void handle_options __P ((int argc, char *argv[]));
+static void handle_options ___P ((int argc, char *argv[]));
 
 /* List options and their values. */
 static void do_list_options ();
@@ -52,7 +61,8 @@ static void version ();
  */
 
 char *program;			/* Program's name, used for messages. */
-FILE *ofp;			/* Output file. */
+FILE *ofp = NULL;		/* Output file. */
+void *printer_context;		/* Context for the printer. */
 char version_string[256];	/* Enscript's version string. */
 char ps_version_string[20];	/* Version string for PS procsets. */
 char date_string[256];		/* Preformatted time string. */
@@ -73,9 +83,6 @@ char afm_path_buffer[1024];
 
 MediaEntry *media_names = NULL;	/* List of known media. */
 MediaEntry *media = NULL;	/* Entry for used media. */
-char spooler_command[256];	/* Name of the spooler command. */
-char queue_param[16];		/* Parameter for print queue (-P). */
-char spooler_options[512];	/* Array of spooler options. */
 int bs = 8;			/* The backspace character. */
 
 /* Statistics. */
@@ -112,6 +119,7 @@ StringHashPtr user_strings;
 
 /* Cache for AFM files. */
 StringHashPtr afm_cache = NULL;
+StringHashPtr afm_info_cache = NULL;
 
 /* AFM library handle. */
 AFMHandle afm = NULL;
@@ -119,6 +127,10 @@ AFMHandle afm = NULL;
 
 /* Options. */
 
+/*
+ * Free single-letter options are: J, Q, w, W, x, y, Y
+ */
+
 /* 
  * -1, -2, --columns
  *
@@ -133,6 +145,14 @@ int num_columns = 1;
  */
 PageRange *page_ranges = NULL;
 
+/*
+ * -A ALIGN, --file-align=ALIGN
+ *
+ * Align input files to start from ALIGN page count.  This is handy
+ * for two-side printings.
+ */
+unsigned int file_align = 1;
+
 /* 
  * -b, --header
  * 
@@ -180,14 +200,28 @@ char printer_buf[256];
 int special_escapes = 0;
 int escape_char = '\0';
 
+/*
+ * -E [LANG], --pretty-print[=LANG]
+ * 
+ * Prettyprint program source code.  Prettyprinting is handled by
+ * creating an input filter with states-preprocessor.  States makes an
+ * educated guess about the start state but sometimes it fails, so the
+ * start state can also be specified to be LANG.  This option
+ * overwrites input filter and enables special escapes.
+ */
+
+int pretty_print = 0;
+char *pp_start_state = NULL;
+char pp_filter[4096];
+
 /*
  * -f, --font
  *
  * Select body font.
  */
 char *Fname = "Courier";
-double Fpt = 10.0;
-double default_Fpt;		/* Point size of the original font. */
+FontPoint Fpt = {10.0, 10.0};
+FontPoint default_Fpt;		/* Point size of the original font. */
 char *default_Fname;		/* Name of the original font. */
 int user_body_font_defined = 0;	/* Has user defined new body font? */
 
@@ -202,7 +236,7 @@ double font_bbox_lly;		/* Font's bounding box's lly-coordinate. */
  * Select font to be used to print the standard simple header.
  */
 char *HFname = "Courier-Bold";
-double HFpt = 10.0;
+FontPoint HFpt = {10.0, 10.0};
 
 /*
  * -g, --print-anyway
@@ -367,7 +401,7 @@ int tabsize = 8;
  * Place text under every page.  Default is no underlay.
  */
 double ul_gray = .8;
-double ul_ptsize = 200.0;
+FontPoint ul_ptsize = {200.0, 200.0};
 char *ul_font = "Times-Roman";
 char *underlay = NULL;
 char ul_position_buf[256];
@@ -381,6 +415,20 @@ char ul_style_str_buf[256];
 int ul_position_p = 0;		/* Is ul-position given? */
 int ul_angle_p = 0;		/* Is ul-angle given? */
 
+/*
+ * -U NUM, --nup=NUM
+ *
+ * Print NUM PostScript pages on each output page (n-up printing).
+ */
+unsigned int nup = 1;
+unsigned int nup_exp = 0;
+unsigned int nup_rows = 1;
+unsigned int nup_columns = 1;
+int nup_landscape = 0;
+unsigned int nup_width;
+unsigned int nup_height;
+double nup_scale;
+
 /*
  * -v, --verbose
  *
@@ -421,6 +469,13 @@ int pass_through = 0;
  */
 char *input_filter_stdin = "";
 
+/*
+ * --help-pretty-print
+ *
+ * Descript all supported -E, --pretty-print languages and file formats.
+ */
+int help_pretty_print = 0;
+
 /*
  * --highlight-bar-gray=val
  *
@@ -467,6 +522,32 @@ PageLabelFormat page_label;
  */
 char *printer_options = NULL;
 
+/*
+ * --slice=NUM
+ *
+ * Horizontal input slicing.  Print only NUMth wrapped input pages.
+ */
+int slicing = 0;
+unsigned int slice = 1;
+
+/*
+ * --toc
+ *
+ * Print Table of Contents page.
+ */
+int toc = 0;
+char toc_fname[512];
+FILE *toc_fp;
+char *toc_fmt_string;
+
+/*
+ * AcceptCompositeCharacters: bool
+ *
+ * Specify whatever we accept composite characters or should them be 
+ * considered as non-existent.  As a default, do not accept them.
+ */
+int accept_composites = 0;
+
 /*
  * AppendCtrlD: bool
  *
@@ -482,23 +563,68 @@ int append_ctrl_D = 0;
  */
 int clean_7bit = 1;
 
-/* ^@shade{GRAY}, set the line highlight gray. */
-double line_highlight_gray = 1.0;
-
 /*
- * AcceptCompositeCharacters, specify whatever we accept composite
- * characters or should them be considered as non-existent.  As a
- * default, do not accept them.
- */
-int accept_composites = 0;
-
-/*
- * FormFeedType, specify what to do when a formfeed character is 
- * encountered from the input stream.  The default action is to jump to
- * the beginning of the next column.
+ * FormFeedType: type
+ * 
+ * Specify what to do when a formfeed character is encountered from the 
+ * input stream.  The default action is to jump to the beginning of the
+ * next column.
  */
 FormFeedType formfeed_type = FORMFEED_COLUMN;
 
+/*
+ * NoJobHeaderSwitch: switch
+ *
+ * Spooler switch to suppress the job header (-h).
+ */
+char no_job_header_switch[16];
+
+/*
+ * OutputFirstLine: line
+ *
+ * Set the PostScript output's first line to something your system can handle.
+ * The default is "%!PS-Adobe-3.0"
+ */
+char output_first_line[256];
+
+/*
+ * QueueParam: param
+ *
+ * The spooler command switch to select the printer queue (-P).
+ */
+char queue_param[16];
+
+/*
+ * Spooler: command
+ *
+ * The spooler command name (lpr).
+ */
+char spooler_command[256];
+
+/*
+ * StatesColorModel: model
+ *
+ * Color model for states.
+ */
+char states_color_model[50];
+
+/*
+ * StatesConfigFile: file
+ *
+ * Read states' configuration from file <file>.
+ */
+char states_config_file[256];
+
+/*
+ * StatesHighlightLevel: level
+ *
+ * HighlightLevel for states.
+ */
+char states_highlight_level[50];
+
+/* ^@shade{GRAY}, set the line highlight gray. */
+double line_highlight_gray = 1.0;
+
 /*
  * Static variables.
  */
@@ -507,6 +633,7 @@ static struct option long_options[] =
 {
   {"columns",		required_argument,	0, 0},
   {"pages",		required_argument,	0, 'a'},
+  {"file-align",	required_argument,	0, 'A'},
   {"header",		required_argument,	0, 'b'},
   {"no-header",		no_argument,		0, 'B'},
   {"truncate-lines",	no_argument,		0, 'c'},
@@ -514,6 +641,7 @@ static struct option long_options[] =
   {"printer",		required_argument,	0, 'd'}, 
   {"setpagedevice",	required_argument,	0, 'D'},
   {"escapes",		optional_argument,	0, 'e'},
+  {"pretty-print",	optional_argument,	0, 'E'},
   {"font",		required_argument,	0, 'f'},
   {"header-font",	required_argument,	0, 'F'},
   {"print-anyway",	no_argument,		0, 'g'},
@@ -542,6 +670,7 @@ static struct option long_options[] =
   {"title",		required_argument,	0, 't'},
   {"tabsize",		required_argument,	0, 'T'},
   {"underlay",		optional_argument,	0, 'u'},
+  {"nup",		required_argument,	0, 'U'},
   {"verbose",		optional_argument,	0, 'v'},
   {"version",		no_argument,		0, 'V'},
   {"encoding",		required_argument,	0, 'X'},
@@ -552,13 +681,15 @@ static struct option long_options[] =
   {"download-font",	required_argument,	0, 131},
   {"filter-stdin",	required_argument,	0, 138},
   {"help", 		no_argument, 		0, 135},
+  {"help-pretty-print", no_argument, 		0, 141},
   {"highlight-bar-gray",	required_argument, 	0, 136},
   {"list-media",	no_argument,		&list_media, 1},
   {"list-options",	no_argument,		&list_options, 1},
-  {"non-printable-format",
-     			required_argument,	0, 134},
+  {"non-printable-format",	required_argument,	0, 134},
   {"page-label-format",	required_argument,	0, 130},
   {"printer-options",	required_argument,	0, 139},
+  {"slice",		required_argument,	0, 140},
+  {"toc",		no_argument,		&toc, 1},
   {"ul-angle",		required_argument,	0, 132},
   {"ul-font",		required_argument,	0, 128},
   {"ul-gray",		required_argument,	0, 129},
@@ -580,12 +711,15 @@ static struct
     {{"latin1", "iso8859", "iso"},	ENC_LATIN1, 		'\n', 8},
     {{"latin2", "iso2", NULL},		ENC_LATIN2, 		'\n', 8},
     {{"latin3", "iso3", NULL},		ENC_LATIN3, 		'\n', 8},
+    {{"latin5", "iso5", NULL},		ENC_LATIN5, 		'\n', 8},
     {{"ascii", NULL, NULL},		ENC_ASCII, 		'\n', 8},
-    {{"asciiscands", NULL, NULL},	ENC_ASCII_SCANDS, 	'\n', 8},
+    {{"asciifise", "asciifi", "asciise"},	ENC_ASCII_FISE,	'\n', 8},
+    {{"asciidkno", "asciidk", "asciino"},	ENC_ASCII_DKNO,	'\n', 8},
     {{"ibmpc", "pc", "dos"},		ENC_IBMPC, 		'\n', 8},
     {{"mac", NULL, NULL},		ENC_MAC, 		'\r', 8},
     {{"vms", NULL, NULL},		ENC_VMS, 		'\n', 8},
     {{"hp8", NULL, NULL},		ENC_HP8,		'\n', 8},
+    {{"koi8", NULL, NULL},		ENC_KOI8,		'\n', 8},
     {{"ps", "PS", NULL},		ENC_PS, 		'\n', 8},
     {{"pslatin1", "ISOLatin1Encoding", NULL},	ENC_LATIN1,	'\n', 8},
 
@@ -604,10 +738,10 @@ main (int argc, char *argv[])
   time_t tim;
   struct tm *tm;
   int i, j, found;
+  unsigned int ui;
   MediaEntry *mentry;
   AFMError afm_error;
   char *cp, *cp2;
-  void *printer_context;
 
   /* Get program's name. */
   program = strrchr (argv[0], '/');
@@ -625,6 +759,9 @@ main (int argc, char *argv[])
   cp = strrchr (ps_version_string, '.');
   *cp = ' ';
 
+  /* Create the default TOC format string.  Wow, this is cool! */
+  toc_fmt_string = _("$3v $-40N $3% pages $4L lines  $E $C");
+
   /* Internationalization. */
 #if HAVE_SETLOCALE 
   /*
@@ -656,12 +793,16 @@ main (int argc, char *argv[])
     fatal (_("couldn't get passwd entry for uid=%d: %s"), getuid (), 
 	   strerror (errno));
 
-  /* Default spooler command name. */
+  /* Defaults for some options. */
   strcpy (spooler_command, "lpr");
   strcpy (queue_param, "-P");
-
-  /* Default fancy header name. */
+  strcpy (no_job_header_switch, "-h");
   strcpy (fancy_header_default, "enscript");
+  strcpy (output_first_line, "%!PS-Adobe-3.0");
+
+  strcpy (states_color_model, "blackwhite");
+  sprintf (states_config_file, "%s/enscript.st", LIBRARY);
+  strcpy (states_highlight_level, "heavy");
 
   /* Check ENSCRIPT_LIBRARY for custom library location. */
   cp = getenv ("ENSCRIPT_LIBRARY");
@@ -759,10 +900,11 @@ main (int argc, char *argv[])
 
   /* Default font for landscape, 2 column printing is Courier 7. */
   if (!user_body_font_defined && landscape && num_columns > 1)
-    Fpt = 7.0;
+    Fpt.w = Fpt.h = 7.0;
 
   /* Cache for font AFM information. */
   afm_cache = strhash_init ();
+  afm_info_cache = strhash_init ();
 
   /* Open AFM library. */
   afm_error = afm_create (afm_path, verbose, &afm);
@@ -778,7 +920,8 @@ main (int argc, char *argv[])
    * Save default Fpt and Fname since special escape 'font' can change 
    * it and later we might want to switch back to the "default" font.
    */
-  default_Fpt = Fpt;
+  default_Fpt.w = Fpt.w;
+  default_Fpt.h = Fpt.h;
   default_Fname = Fname;
 
   /* Register that document uses at least these fonts. */
@@ -843,6 +986,62 @@ name             width\theight\tllx\tlly\turx\tury\n\
   else
     fatal (_("illegal non-printable format \"%s\""), npf_name);
 
+  /* N-up. */
+  for (i = 0; ; i++)
+    {
+      ui = nup >> i;
+
+      if (ui == 0)
+	fatal (_("illegal N-up argument: %d"), nup);
+
+      if (ui & 0x1)
+	{
+	  if (ui != 1)
+	    fatal (_("N-up argument must be power of 2: %d"), nup);
+	  
+	  nup_exp = i;
+	  break;
+	}
+    }
+
+  nup_rows = nup_exp / 2 * 2;
+  if (nup_rows == 0)
+    nup_rows = 1;
+  nup_columns = (nup_exp + 1) / 2 * 2;
+  if (nup_columns == 0)
+    nup_columns = 1;
+
+  nup_landscape = nup_exp & 0x1;
+
+  if (nup_landscape)
+    {
+      nup_width = media->h;
+      nup_height = media->w;
+    }
+  else
+    {
+      nup_width = media->w;
+      nup_height = media->h;
+    }
+  nup_scale = (double) nup_width / nup_columns / media->w;
+  nup_width /= nup_columns;
+  nup_height /= nup_rows;
+
+  /*
+   * We print the data inside media's bouding box but the data has also
+   * its own marginals => they get printed 1.n times so we must adjust
+   * marginals a bit to get them right.
+   */
+  if (nup > 1)
+    {
+      double adjscale = 1.0 / (1 + nup_scale);
+
+      media->llx *= adjscale;
+      media->lly *= adjscale;
+      media->urx = media->w - ((media->w - media->urx) * adjscale);
+      media->ury = media->h - ((media->h - media->ury) * adjscale);
+    }
+
   /* 
    * Count output media dimensions.
    */
@@ -911,7 +1110,7 @@ name             width\theight\tllx\tlly\turx\tury\n\
       break;
 
     case HDR_SIMPLE:
-      d_header_h = HFpt * 1.5;
+      d_header_h = HFpt.h * 1.5;
       break;
 
     case HDR_FANCY:
@@ -919,6 +1118,32 @@ name             width\theight\tllx\tlly\turx\tury\n\
       break;
     }
 
+  /* Help pretty-print. */
+  if (help_pretty_print)
+    {
+      /* Create description with states. */
+      sprintf (pp_filter,
+	       "states -f %s -s describe_languages %s",
+	       states_config_file, states_config_file);
+      system (pp_filter);
+      exit (0);
+    }
+
+  /* Pretty-printing */
+  if (pretty_print)
+    {
+      /* Create pretty-print input filter. */
+      sprintf (pp_filter,
+	       "states -f %s %s%s -Dcolormodel=%s -Dhl_level=%s -Dfont_spec=%s@%g/%g %%s",
+	       states_config_file,
+	       pp_start_state ? "-s " : "",
+	       pp_start_state ? pp_start_state : "",
+	       states_color_model,
+	       states_highlight_level,
+	       Fname, Fpt.w, Fpt.h);
+      input_filter = pp_filter;
+    }
+
   /* List options. */
   if (list_options)
     {
@@ -926,32 +1151,21 @@ name             width\theight\tllx\tlly\turx\tury\n\
       exit (1);
     }
 
-  /* Open output file. */
-  if (output_file == OUTPUT_FILE_NONE)
+  /* Table of Contents. */
+  if (toc)
     {
-      /* Format spooler options. */
-      spooler_options[0] = '\0';
-      if (mail)
-	strcat (spooler_options, "-m ");
-      if (no_job_header)
-	strcat (spooler_options, "-h ");
-      if (printer_options)
-	strcat (spooler_options, printer_options);
-
-      /* Open printer. */
-      ofp = printer_open (spooler_command, spooler_options, queue_param,
-			  printer, &printer_context);
-      if (ofp == NULL)
-	fatal (_("couldn't open printer `%s': %s"), printer, strerror (errno));
-    }
-  else if (output_file == OUTPUT_FILE_STDOUT)
-    ofp = stdout;
-  else
-    {
-      ofp = fopen (output_file, "w");
-      if (ofp == NULL)
-	fatal (_("couldn't create output file \"%s\": %s"), output_file,
+      cp = tmpnam (toc_fname);
+      if (cp == NULL)
+	fatal (_("couldn't create toc file name: %s"), strerror (errno));
+      
+      toc_fp = fopen (toc_fname, "w+b");
+      if (toc_fp == NULL)
+	fatal (_("couldn't create toc file \"%s\": %s"), toc_fname,
 	       strerror (errno));
+      
+      if (remove (toc_fname) == 0)
+	/* Remove successfull, no need to remove file at exit. */
+	toc_fname[0] = '\0';
     }
 
 
@@ -966,6 +1180,8 @@ name             width\theight\tllx\tlly\turx\tury\n\
 
       if (is_open (&is, stdin, NULL, input_filter))
 	{
+	  /* Open output file. */
+	  open_output_file ();
 	  process_file (title_given ? title : "", &is);
 	  is_close (&is);
 	}
@@ -985,6 +1201,13 @@ name             width\theight\tllx\tlly\turx\tury\n\
 		  tm = localtime (&tim);
 		  memcpy (&mod_tm, tm, sizeof (*tm));
 
+		  /*
+		   * Open output file.  Output file opening is delayed to
+		   * this point so we can optimize the case when a 
+		   * non-existing input file is printed => we do nothing.
+		   */
+		  open_output_file ();
+
 		  process_file (argv[optind], &is);
 		}
 	      else
@@ -996,6 +1219,33 @@ name             width\theight\tllx\tlly\turx\tury\n\
 	}
     }
 
+  /* Table of Contents. */
+  if (toc)
+    {
+      /* This is really cool... */
+
+      /* Set the printing options for toc. */
+      toc = 0;
+      special_escapes = 1;
+
+      if (fseek (toc_fp, 0, SEEK_SET) != 0)
+	fatal (_("couldn't rewind toc file: %s"), strerror (errno));
+
+      memcpy (&mod_tm, &run_tm, sizeof (run_tm));
+      if (is_open (&is, toc_fp, NULL, NULL))
+	{
+	  process_file (_("Table of Contents"), &is);
+	  is_close (&is);
+	}
+
+      /* Clean up toc file. */
+      fclose (toc_fp);
+      
+      /* Do we have to remove the toc file? */
+      if (toc_fname[0])
+	(void) remove (toc_fname);
+    }
+
   /* Give trailer a chance to dump itself. */
   dump_ps_trailer ();
 
@@ -1003,43 +1253,65 @@ name             width\theight\tllx\tlly\turx\tury\n\
    * Append ^D to the end of the output?  Note! It must be ^D followed
    * by a newline.
    */
-  if (append_ctrl_D)
+  if (ofp != NULL && append_ctrl_D)
     fprintf (ofp, "\004\n");
 
   /* Close output file. */
-  if (output_file == OUTPUT_FILE_NONE)
-    printer_close (printer_context);
-  else if (output_file != OUTPUT_FILE_STDOUT)
-    fclose (ofp);
+  close_output_file ();
 
   /* Tell how things went. */
-  message (0, _("[ %d pages * %d copy ]"), total_pages, num_copies);
-  if (output_file == OUTPUT_FILE_NONE)
-    message (0, _(" sent to %s\n"), printer ? printer : _("printer"));
-  else
-    message (0, _(" left in %s\n"),
-	     output_file == OUTPUT_FILE_STDOUT ? "-" : output_file);
-  if (num_truncated_lines)
-    message (0, _("%d lines were %s\n"), num_truncated_lines,
-	     truncate_lines ? _("truncated") : _("wrapped"));
-
-  if (num_missing_chars)
+  if (ofp == NULL)
     {
-      message (0, _("%d characters were missing\n"), num_missing_chars);
-      if (list_missing_characters)
-	{
-	  message (0, _("missing character codes (decimal):\n"));
-	  do_list_missing_characters (missing_chars);
-	}
+      /*
+       * The value of <ofp> is not reset in close_output_file(),
+       * this is ugly but it saves one flag.
+       */
+      message (0, _("no output generated\n"));
     }
-
-  if (num_non_printable_chars)
+  else
     {
-      message (0, _("%d non-printable characters\n"), num_non_printable_chars);
-      if (list_missing_characters)
+      unsigned int real_total_pages;
+
+      if (nup > 1)
 	{
-	  message (0, _("non-printable character codes (decimal):\n"));
-	  do_list_missing_characters (non_printable_chars);
+	  if (total_pages > 0)
+	    real_total_pages = (total_pages - 1) / nup + 1;
+	  else
+	    real_total_pages = 0;
+	}
+      else
+	real_total_pages = total_pages;
+
+      /* We did something, tell what.  */
+      message (0, _("[ %d pages * %d copy ]"), real_total_pages, num_copies);
+      if (output_file == OUTPUT_FILE_NONE)
+	message (0, _(" sent to %s\n"), printer ? printer : _("printer"));
+      else
+	message (0, _(" left in %s\n"),
+		 output_file == OUTPUT_FILE_STDOUT ? "-" : output_file);
+      if (num_truncated_lines)
+	message (0, _("%d lines were %s\n"), num_truncated_lines,
+		 truncate_lines ? _("truncated") : _("wrapped"));
+
+      if (num_missing_chars)
+	{
+	  message (0, _("%d characters were missing\n"), num_missing_chars);
+	  if (list_missing_characters)
+	    {
+	      message (0, _("missing character codes (decimal):\n"));
+	      do_list_missing_characters (missing_chars);
+	    }
+	}
+
+      if (num_non_printable_chars)
+	{
+	  message (0, _("%d non-printable characters\n"),
+		   num_non_printable_chars);
+	  if (list_missing_characters)
+	    {
+	      message (0, _("non-printable character codes (decimal):\n"));
+	      do_list_missing_characters (non_printable_chars);
+	    }
 	}
     }
 
@@ -1052,6 +1324,62 @@ name             width\theight\tllx\tlly\turx\tury\n\
  * Static functions.
  */
 
+static void
+open_output_file ()
+{
+  if (ofp)
+    /* Output file has already been opened, do nothing. */
+    return;
+
+  if (output_file == OUTPUT_FILE_NONE)
+    {
+      char spooler_options[512];
+
+      /* Format spooler options. */
+      spooler_options[0] = '\0';
+      if (mail)
+	strcat (spooler_options, "-m ");
+      if (no_job_header)
+	{
+	  strcat (spooler_options, no_job_header_switch);
+	  strcat (spooler_options, " ");
+	}
+      if (printer_options)
+	strcat (spooler_options, printer_options);
+
+      /* Open printer. */
+      ofp = printer_open (spooler_command, spooler_options, queue_param,
+			  printer, &printer_context);
+      if (ofp == NULL)
+	fatal (_("couldn't open printer `%s': %s"), printer, strerror (errno));
+    }
+  else if (output_file == OUTPUT_FILE_STDOUT)
+    ofp = stdout;
+  else
+    {
+      ofp = fopen (output_file, "w");
+      if (ofp == NULL)
+	fatal (_("couldn't create output file \"%s\": %s"), output_file,
+	       strerror (errno));
+    }
+}
+
+
+static void
+close_output_file ()
+{
+  if (ofp == NULL)
+    /* Output file hasn't been opened, we are done. */
+    return;
+
+  if (output_file == OUTPUT_FILE_NONE)
+    printer_close (printer_context);
+  else if (output_file != OUTPUT_FILE_STDOUT)
+    fclose (ofp);
+
+  /* We do not reset <ofp> since its value is needed in diagnostigs. */
+}
+
 
 static void
 handle_env_options (char *var)
@@ -1066,7 +1394,7 @@ handle_env_options (char *var)
   if (string == NULL)
     return;
 
-  message (1, "handle_env_options(): %s=\"%s\"\n", var, string);
+  message (2, "handle_env_options(): %s=\"%s\"\n", var, string);
 
   /* Copy string so we can modify it in place. */
   str = xstrdup (string);
@@ -1167,7 +1495,7 @@ handle_options (int argc, char *argv[])
       const char *cp;
 
       c = getopt_long (argc, argv,
-		       "12a:b:BcCd:D:e::f:F:gGhH::i:I:jkKlL:mM:n:N:o:Op:P:qrRs:S:t:T:u::vVX:zZ",
+		       "12a:A:b:BcCd:D:e::E::f:F:gGhH::i:I:jkKlL:mM:n:N:o:Op:P:qrRs:S:t:T:u::U:vVX:zZ",
 		       long_options, &option_index);
       
       if (c == EOF)
@@ -1226,6 +1554,12 @@ handle_options (int argc, char *argv[])
 	  page_ranges = prange;
 	  break;
 
+	case 'A':		/* file alignment */
+	  file_align = atoi (optarg);
+	  if (file_align == 0)
+	    fatal (_("file alignment must be larger than zero"));
+	  break;
+
 	case 'b':		/* page header */
 	  page_header = optarg;
 	  break;
@@ -1259,6 +1593,13 @@ handle_options (int argc, char *argv[])
 	    escape_char = atoi (optarg);
 	  break;
 
+	case 'E':		/* pretty-print */
+	  pretty_print = 1;
+	  special_escapes = 1;
+	  escape_char = '\0';
+	  pp_start_state = optarg;
+	  break;
+
 	case 'f':		/* font */
 	  if (!parse_font_spec (optarg, &Fname, &Fpt))
 	    fatal (_("malformed font spec: %s"), optarg);
@@ -1406,6 +1747,10 @@ handle_options (int argc, char *argv[])
 	  underlay = optarg;
 	  break;
 
+	case 'U':		/* nup */
+	  nup = atoi (optarg);
+	  break;
+
 	case 'v':		/* verbose */
 	  if (optarg)
 	    verbose = atoi (optarg);
@@ -1484,6 +1829,17 @@ handle_options (int argc, char *argv[])
 	  printer_options = optarg;
 	  break;
 
+	case 140:		/* slicing */
+	  slicing = 1;
+	  slice = atoi (optarg);
+	  if (slice <= 0)
+	    fatal (_("slice must be greater than zero"));
+	  break;
+
+	case 141:		/* help-pretty-print */
+	  help_pretty_print = 1;
+	  break;
+
 	case '?':		/* Errors found during getopt_long(). */
 	option_error:
 	  fprintf (stderr, _("Try `%s --help' for more information.\n"),
@@ -1542,8 +1898,8 @@ do_list_options ()
   printf ("\n");
   
   printf ("page_header=\"%s\"\n", page_header ? page_header : "");
-  printf ("font: name=%s size=%gpt\n", Fname, Fpt);
-  printf ("header font: name=%s size=%gpt\n", HFname, HFpt);
+  printf ("font: name=%s size=%g/%gpt\n", Fname, Fpt.w, Fpt.h);
+  printf ("header font: name=%s size=%g/%gpt\n", HFname, HFpt.w, HFpt.h);
   printf ("output_file=%s\n",
 	  (output_file == OUTPUT_FILE_NONE
 	   ? "none"
@@ -1582,7 +1938,7 @@ do_list_options ()
   /* Underlay. */
   printf ("underlay=(%s)\n", 	underlay ? underlay : "");
   printf ("ul_gray=%g\n", 		ul_gray);
-  printf ("ul_font=%s %gpt\n", 	ul_font, ul_ptsize);
+  printf ("ul_font=%s %g/%gpt\n", 	ul_font, ul_ptsize.w, ul_ptsize.h);
   printf ("ul_position=%s (%g, %g)\n",	ul_position, ul_x, ul_y);
   printf ("ul_angle=%g\n",		ul_angle);
   
@@ -1605,6 +1961,7 @@ Mandatory arguments to long options are mandatory for short options too.\n\
   -2                         same as --columns=2\n\
       --columns=NUM          specify the number of columns per page\n\
   -a, --pages=PAGES          specify which pages are printed\n\
+  -A, --file-align=ALIGN     align separate input files to ALIGN\n\
   -b, --header=HEADER        set page header\n\
   -B, --no-header            no page headers\n\
   -c, --truncate-lines       cut long lines (default is to wrap)\n\
@@ -1615,6 +1972,9 @@ Mandatory arguments to long options are mandatory for short options too.\n\
   -e, --escapes[=CHAR]       enable special escape interpretation\n"),
           program);
 
+  printf (_("\
+  -E, --pretty-print[=LANG]  pretty-print source code\n"));
+
   printf (_("\
   -f, --font=NAME            use font NAME for body text\n\
   -F, --header-font=NAME     use font NAME for header texts\n\
@@ -1656,6 +2016,7 @@ Mandatory arguments to long options are mandatory for short options too.\n\
                              sets also the name of the input file stdin.\n\
   -T, --tabsize=NUM          set tabulator size to NUM\n\
   -u, --underlay[=TEXT]      print TEXT under every page\n\
+  -U, --nup=NUM              print NUM logical pages on each output page\n\
   -v, --verbose              tell what we are doing\n\
   -V, --version              print version number\n\
   -X, --encoding=NAME        use input encoding NAME\n\
@@ -1667,12 +2028,18 @@ Mandatory arguments to long options are mandatory for short options too.\n\
   --download-font=NAME       download font NAME\n\
   --filter-stdin=NAME        specify how stdin is shown to the input filter\n\
   --help                     print this help and exit\n\
+  --help-pretty-print        describe all supported --pretty-print languages\n\
+                             and file formats\n\
   --highlight-bar-gray=NUM   print highlight bars with gray NUM (0 - 1)\n\
   --list-media               list names of all known media\n\
   --list-options             list all options and their values\n\
   --non-printable-format=FMT specify how non-printable chars are printed\n\
   --page-label-format=FMT    set page label format to FMT\n\
-  --printer-options=OPTIONS  pass extra options to the printer command\n\
+  --printer-options=OPTIONS  pass extra options to the printer command\n"));
+  
+  printf (_("\
+  --slice=NUM                print vertical slice NUM\n\
+  --toc                      print table of contents\n\
   --ul-angle=ANGLE           set underlay text's angle to ANGLE\n\
   --ul-font=NAME             print underlays with font NAME\n\
   --ul-gray=NUM              print underlays with gray value NUM\n\
