@@ -47,6 +47,9 @@
 #define F_EPSF_SCALE_Y			0x80
 
 
+/* Predicate to check if we are at the correct slice. */
+#define CORRECT_SLICE() (slicing == 0 || current_slice == slice)
+
 /* Predicates for the current body font. */
 
 /* Is character <ch> printable. */
@@ -73,6 +76,7 @@ typedef enum
   tSETPAGENUMBER,
   tNEWPAGE,
   tFONT,
+  tCOLOR,
   tPS
 } TokenType;
 
@@ -82,6 +86,7 @@ typedef enum
   ESC_COMMENT,
   ESC_EPSF,
   ESC_FONT,
+  ESC_COLOR,
   ESC_NEWPAGE,
   ESC_SETFILENAME,
   ESC_SETPAGENUMBER,
@@ -118,10 +123,11 @@ struct gs_token_st
 	  FILE *fp;		/* File from which eps image is read. */
 	  int pipe;		/* Is <fp> opened to pipe?  */
 	} epsf;
+      Color color;
       struct
 	{
 	  char name[512];
-	  double size;
+	  FontPoint size;
 	} font;
       char filename[512];
     } u;
@@ -134,33 +140,35 @@ typedef struct gs_token_st Token;
  * Prototypes for static functions.
  */
 
-static void get_next_token __P ((InputStream *is, double linepos,
-				 unsigned int col, double linew,
-				 Token *token));
+static void get_next_token ___P ((InputStream *is, double linepos,
+				  unsigned int col, double linew,
+				  Token *token));
 
-static void dump_ps_page_header __P ((char *fname));
+static void dump_ps_page_header ___P ((char *fname));
 
 static void dump_ps_page_trailer ();
 
+static void dump_empty_page ();
+
 /*
  * Recognize a EPS file described by <token>.  Returns 1 if file was a
  * valid EPS file or 0 otherwise.  File is accepted if it starts with
  * the PostScript magic `%!' and it has a valid `%%BoundingBox' DSC
  * comment.
  */
-static int recognize_eps_file __P ((Token *token));
+static int recognize_eps_file ___P ((Token *token));
 
 /*
  * Insert EPS file described by <token> to the output stream.
  */
-static void paste_epsf __P ((Token *token));
+static void paste_epsf ___P ((Token *token));
 
 /*
  * Check if InputStream <is> contains a file which can be passed
  * through without any modifications.  Returns 1 if file was passed or
  * 0 otherwise.
  */
-static int do_pass_through __P ((char *fname, InputStream *is));
+static int do_pass_through ___P ((char *fname, InputStream *is));
 
 /*
  * Read one float dimension from InputStream <is>.  If <units> is
@@ -168,17 +176,17 @@ static int do_pass_through __P ((char *fname, InputStream *is));
  * <horizontal> is true, dimension is horizontal, otherwise it is
  * vertical (this is used to find out how big `line' units are).
  */
-static double read_float __P ((InputStream *is, int units, int horizontal));
+static double read_float ___P ((InputStream *is, int units, int horizontal));
 
 /*
  * Print linenumber <linenum> to the beginning of the current line.
  * Current line start is specified by point (x, y).
  */
-static void print_line_number __P ((double x, double y, double space,
-				    double margin, unsigned int linenum));
+static void print_line_number ___P ((double x, double y, double space,
+				     double margin, unsigned int linenum));
 
 /* Send PostScript to the output file. */
-static void output __P ((char *fmt, ...));
+static void output ___P ((char *fmt, ...));
 
 /* Divert output to tmp file so the total page count can be counted. */
 static void divert ();
@@ -191,9 +199,11 @@ static void undivert ();
  * Global variables.
  */
 
-/* The number of the current page. */
-unsigned int current_pagenum;
+unsigned int current_pagenum;	/* The number of the current page. */
 unsigned int total_pages_in_file;
+unsigned int input_filenum = 0;
+unsigned int current_file_linenum;
+char fname[1024];		/* The name of the current input file. */
 
 
 /*
@@ -213,11 +223,17 @@ static FILE *cofp = NULL;
 static int do_print = 1;
 
 /* Is ^@font{}-defined font active? */
-static int user_font = 0;
+static int user_fontp = 0;
 
 /* The user ^@font{}-defined font. */
 static char user_font_name[256];
-static double user_font_pt;
+static FontPoint user_font_pt;
+
+/* Is ^@color{}-defined color active? */
+static int user_colorp = 0;
+
+/* The user ^@color{}-defined color. */
+static Color user_color;
 
 /* The last linenumber printed by print_line_number(). */
 static unsigned int print_line_number_last;
@@ -241,14 +257,16 @@ dump_ps_header ()
    * Header.
    */
 
-  output ("%%!PS-Adobe-3.0\n");
+  output ("%s\n", output_first_line);
   output ("%%%%BoundingBox: %d %d %d %d\n", media->llx, media->lly,
 	  media->urx, media->ury);
   output ("%%%%Title: %s\n", title);
   output ("%%%%For: %s\n", passwd->pw_gecos);
   output ("%%%%Creator: %s\n", version_string);
   output ("%%%%CreationDate: %s\n", date_string);
-  output ("%%%%Orientation: %s\n", landscape ? "Landscape" : "Portrait");
+  output ("%%%%Orientation: %s\n",
+	  ((nup > 1) && nup_landscape)
+	  || ((nup == 1) && landscape) ? "Landscape" : "Portrait");
   output ("%%%%Pages: (atend)\n");
   output ("%%%%DocumentMedia: %s %d %d 0 () ()\n",
 	  media->name, media->w, media->h);
@@ -310,31 +328,35 @@ dump_ps_header ()
   else
     output ("/user_header_p false def\n");
 
-  output ("/HFpt %g def\n", HFpt);
+  output ("/HFpt_w %g def\n", HFpt.w);
+  output ("/HFpt_h %g def\n", HFpt.h);
 
 
   /* Select our fonts. */
 
   /* Header font HF. */
   output ("/%s /HF-gs-font MF\n", HFname);
-  output ("/HF /HF-gs-font findfont HFpt scalefont def\n");
+  output ("/HF /HF-gs-font findfont [HFpt_w 0 0 HFpt_h 0 0] makefont def\n");
 
   /* Our default typing font F. */
   output ("/%s /F-gs-font MF\n", Fname);
-  output ("/F-gs-font findfont %g scalefont setfont\n", Fpt);
+  output ("/F-gs-font findfont [%g 0 0 %g 0 0] makefont setfont\n",
+	  Fpt.w, Fpt.h);
 
   /* Underlay. */
   if (underlay != NULL)
     {
       output ("/ul_str (%s) def\n", underlay);
-      output ("/ul_ptsize %g def\n", ul_ptsize);
+      output ("/ul_w_ptsize %g def\n", ul_ptsize.w);
+      output ("/ul_h_ptsize %g def\n", ul_ptsize.h);
       output ("/ul_gray %g def\n", ul_gray);
       output ("/ul_x %g def\n", ul_x);
       output ("/ul_y %g def\n", ul_y);
       output ("/ul_angle %g def\n", ul_angle);
       output ("/ul_style %d def\n", ul_style);
       output ("/%s /F-ul-font MF\n", ul_font);
-      output ("/ul_font /F-ul-font findfont ul_ptsize scalefont def\n");
+      output ("/ul_font /F-ul-font findfont \
+[ul_w_ptsize 0 0 ul_h_ptsize 0 0] makefont def\n");
     }
 
   /* Number of copies. */
@@ -440,11 +462,21 @@ dump_ps_trailer ()
   int i, j, got;
   char *cp;
   void *value;
+  unsigned int nup_subpage;
 
   if (!ps_header_dumped)
     /* No header, let's be consistent and forget trailer also. */
     return;
 
+  /* The possible pending N-up showpage. */
+  nup_subpage = (total_pages - 1) % nup;
+  if (nup > 1 && nup_subpage + 1 != nup)
+    {
+      /* N-up showpage missing. */
+      output ("_R\n");
+      output ("S\n");
+    }
+
   /* Trailer. */
 
   output ("%%%%Trailer\n");
@@ -488,19 +520,24 @@ process_file (char *fname_arg, InputStream *is)
   int page_clear = 1;
   unsigned int line_column;
   unsigned int current_linenum;
-  unsigned int current_file_linenum = 1;
   double linenumber_space = 0;
   double linenumber_margin = 0;
   Token token;
   int reuse_last_token = 0;
-  char fname[512];
+  unsigned int current_slice = 1;
+  int last_wrapped_line = -1;
+  int last_spaced_file_linenum = -1;
 
   /* Save filename. */
   strcpy (fname, fname_arg);
 
-  /* Init page number count. */
+  /* Init page number and line counters. */
   current_pagenum = 0;
   total_pages_in_file = 0;
+  current_file_linenum = 1;
+
+  /* We got a new input file. */
+  input_filenum++;
 
   /* We haven't printed any line numbers yet. */
   print_line_number_last = (unsigned int) -1;
@@ -511,6 +548,16 @@ process_file (char *fname_arg, InputStream *is)
   /* We have work to do, let's give header a chance to dump itself. */
   dump_ps_header ();
 
+  /*
+   * Align files to the file_align boundary, this is handy for two-side
+   * printing.
+   */
+  while ((total_pages % file_align) != 0)
+    {
+      dump_empty_page ();
+      total_pages++;
+    }
+
   message (1, _("processing file \"%s\"...\n"), fname);
 
   linewidth = d_output_w / num_columns - 2 * d_output_x_margin
@@ -536,7 +583,8 @@ process_file (char *fname_arg, InputStream *is)
 
 	  while (1)
 	    {
-	      if (line_numbers && line_column == 0)
+	      if (line_numbers && line_column == 0
+		  && (current_file_linenum != last_spaced_file_linenum))
 		{
 		  /* 
 		   * Count the space needed by linenumbers.  This should
@@ -548,6 +596,7 @@ process_file (char *fname_arg, InputStream *is)
 
 		  /* Forward x by the amount needed by our line numbers. */
 		  x += linenumber_space + linenumber_margin;
+		  last_spaced_file_linenum = current_file_linenum;
 		}
 
 	      /* Get token. */
@@ -626,8 +675,8 @@ process_file (char *fname_arg, InputStream *is)
 	      if (line_column == 0 && line_highlight_gray < 1.0)
 		output ("%g %g %g %g %g line_highlight\n",
 			lx, (y - baselineskip
-			     + (font_bbox_lly * Fpt / UNITS_PER_POINT)),
-			linewidth, Fpt + baselineskip, line_highlight_gray);
+			     + (font_bbox_lly * Fpt.h / UNITS_PER_POINT)),
+			linewidth, Fpt.h + baselineskip, line_highlight_gray);
 
 	      /* Check rest of tokens. */
 	      switch (token.type)
@@ -646,8 +695,11 @@ process_file (char *fname_arg, InputStream *is)
 		  break;
 
 		case tSTRING:
-		  output ("%g %g M\n", x, y);
-		  output ("(%s) s\n", token.u.str);
+		  if (CORRECT_SLICE ())
+		    {
+		      output ("%g %g M\n", x, y);
+		      output ("(%s) s\n", token.u.str);
+		    }
 		  x = token.new_x;
 		  line_column = token.new_col;
 		  break;
@@ -660,12 +712,33 @@ process_file (char *fname_arg, InputStream *is)
 		  break;
 		  
 		case tNEWLINE:
-		  current_file_linenum++;
-		  /* FALLTHROUGH */
-
 		case tWRAPPED_NEWLINE:
+		  if (token.type == tNEWLINE)
+		    {
+		      current_file_linenum++;
+		      current_slice = 1;
+		      y -= LINESKIP;
+		    }
+		  else
+		    {
+		      current_slice++;
+		      if (!slicing)
+			/*
+			 * For wrapped newlines, decrement y only if
+			 * we are not slicing the input.
+			 */
+			y -= LINESKIP;
+
+		      /* Count the wrapped lines here. */
+		      if (!slicing || current_slice > slice)
+			if (current_file_linenum != last_wrapped_line)
+			  {
+			    num_truncated_lines++;
+			    last_wrapped_line = current_file_linenum;
+			  }
+		    }
+
 		  current_linenum++;
-		  y -= LINESKIP;
 		  if (current_linenum >= lines_per_page
 		      || y < d_footer_h + d_output_y_margin)
 		    goto end_of_column;
@@ -721,7 +794,8 @@ large for page\n"),
 		    }
 
 		  /* Do paste. */
-		  paste_epsf (&token);
+		  if (CORRECT_SLICE ())
+		    paste_epsf (&token);
 
 		  /* Update current point? */
 		  if (!(token.flags & F_EPSF_NO_CPOINT_UPDATE_Y))
@@ -735,15 +809,17 @@ large for page\n"),
 
 		case tFONT:
 		  /* Select a new current font. */
-		  message (1, "^@font=");
+		  message (2, "^@font=");
 		  if (token.u.font.name[0] == '\0')
 		    {
 		      /* Select the default font. */
-		      Fpt = default_Fpt;
+		      Fpt.w = default_Fpt.w;
+		      Fpt.h = default_Fpt.h;
 		      Fname = default_Fname;
-		      output ("/F-gs-font findfont %g scalefont setfont\n",
-			      Fpt);
-		      user_font = 0;
+		      output ("/F-gs-font findfont \
+[%g 0 0 %g 0 0] makefont setfont\n",
+			      Fpt.w, Fpt.h);
+		      user_fontp = 0;
 		    }
 		  else
 		    {
@@ -752,21 +828,51 @@ large for page\n"),
 				   NULL, NULL);
 		      output ("/%s /F-gs-user-font MF\n",
 			      token.u.font.name);
-		      output ("/F-gs-user-font findfont %g scalefont \
-setfont\n",
-			       token.u.font.size);
+		      output ("/F-gs-user-font findfont \
+[%g 0 0 %g 0 0] makefont setfont\n",
+			      token.u.font.size.w, token.u.font.size.h);
 
 		      strcpy (user_font_name, token.u.font.name);
-		      user_font_pt = token.u.font.size;
-		      user_font = 1;
+		      user_font_pt.w = token.u.font.size.w;
+		      user_font_pt.h = token.u.font.size.h;
+		      user_fontp = 1;
 		      
-		      Fpt = user_font_pt;
+		      Fpt.w = user_font_pt.w;
+		      Fpt.h = user_font_pt.h;
 		      Fname = user_font_name;
 		    }
-		  message (1, "%s %gpt\n", Fname, Fpt);
+		  message (2, "%s %g/%gpt\n", Fname, Fpt.w, Fpt.h);
 		  read_font_info ();
 		  break;
 
+		case tCOLOR:
+		  /* Select a new color. */
+		  message (2, "^@color{%f %f %f}\n",
+			   token.u.color.r,
+			   token.u.color.g,
+			   token.u.color.b);
+		  if (token.u.color.r == token.u.color.g
+		      && token.u.color.g == token.u.color.b
+		      && token.u.color.b == 0.0)
+		    {
+		      /* Select the default color (black). */
+		      output ("0 setgray\n");
+		      user_colorp = 0;
+		    }
+		  else
+		    {
+		      output ("%g %g %g setrgbcolor\n",
+			      token.u.color.r,
+			      token.u.color.g,
+			      token.u.color.b);
+
+		      user_color.r = token.u.color.r;
+		      user_color.g = token.u.color.g;
+		      user_color.b = token.u.color.b;
+		      user_colorp = 1;
+		    }
+		  break;
+
 		case tSETFILENAME:
 		  strcpy (fname, token.u.filename);
 		  break;
@@ -808,6 +914,16 @@ setfont\n",
   do_print = 1;
   
   undivert ();
+  
+  /* Table of contents? */
+  if (toc)
+    {
+      char *cp;
+
+      cp = format_user_string ("TOC", toc_fmt_string);
+      fprintf (toc_fp, "%s\n", cp);
+      xfree (cp);
+    }
 }
 
 
@@ -857,6 +973,7 @@ static struct
     {"comment",		ESC_COMMENT},
     {"epsf", 		ESC_EPSF},
     {"font", 		ESC_FONT},
+    {"color",		ESC_COLOR},
     {"newpage",		ESC_NEWPAGE},
     {"ps",		ESC_PS},
     {"setfilename",	ESC_SETFILENAME},
@@ -1141,6 +1258,42 @@ read_special_escape (InputStream *is, Token *token)
 	  token->type = tFONT;
 	  break;
 
+	case ESC_COLOR:
+	  /* Check for the default color. */
+	  if (strcmp (buf, "default") == 0)
+	    {
+	      token->u.color.r = 0;
+	      token->u.color.g = 0;
+	      token->u.color.b = 0;
+	    }
+	  else
+	    {
+	      int got;
+
+	      got = sscanf (buf, "%g %g %g",
+			    &token->u.color.r,
+			    &token->u.color.g,
+			    &token->u.color.b);
+	      switch (got)
+		{
+		case 0:
+		case 2:
+		  fatal (_("malformed color spec for ^@color escape: %s"),
+			 buf);
+		  break;
+
+		case 1:
+		  token->u.color.g = token->u.color.b = token->u.color.r;
+		  break;
+
+		default:
+		  /* Got all three components. */
+		  break;
+		}
+	    }
+	  token->type = tCOLOR;
+	  break;
+
 	case ESC_SHADE:
 	  line_highlight_gray = atof (buf);
 	  if (line_highlight_gray < 0.0 || line_highlight_gray > 1.0)
@@ -1468,8 +1621,6 @@ get_next_token (InputStream *is, double linepos, unsigned int col,
   if (done == DONE_WRAP)
     {
       /* This line is too long. */
-      num_truncated_lines++;
-
       ch = nl;
       if (truncate_lines)
 	{
@@ -1503,6 +1654,11 @@ dump_ps_page_header (char *fname)
   char *ftail;
   int got, i;
   char *cp, *cp2;
+  char *cstr = "%%";
+  unsigned int nup_subpage;
+
+  /* The N-up printing sub-page. */
+  nup_subpage = (total_pages - 1) % nup;
 
   /* Create fdir and ftail. */
   ftail = strrchr (fname, '/');
@@ -1517,16 +1673,81 @@ dump_ps_page_header (char *fname)
       strncpy (buf, fname, ftail - fname);
       buf[ftail - fname] = '\0';
     }
+  
+  if (nup > 1)
+    {
+      /* N-up printing is active. */
+      cstr = "%";
+
+      if (nup_subpage == 0)
+	{
+	  /* This is a real page start. */
+
+	  switch (page_label)
+	    {
+	    case LABEL_SHORT:
+	      output ("%%%%Page: (%d-%d) %d\n", current_pagenum,
+		      current_pagenum + nup - 1, total_pages / nup + 1);
+	      break;
+
+	    case LABEL_LONG:
+	      output ("%%%%Page: (%s:%3d-%3d) %d\n", ftail, current_pagenum,
+		      current_pagenum + nup - 1, total_pages / nup + 1);
+	      break;
+	    }
+
+	  /* Page setup. */
+	  output ("%%%%BeginPageSetup\n");
+	  output ("_S\n");
+
+#define PRINT_BOUNDING_BOXES 0
+
+#if PRINT_BOUNDING_BOXES
+	  output ("%d %d moveto %d %d lineto %d %d lineto %d %d lineto closepath stroke\n",
+		  media->llx, media->lly, media->llx, media->ury,
+		  media->urx, media->ury, media->urx, media->lly);
+#endif
+
+	  if (landscape)
+	    {
+	      if (nup_landscape)
+		{
+		  output ("90 rotate\n");
+		  output ("%d %d translate\n", media->lly, -media->urx);
+		}
+	      else
+		output ("%d %d translate\n", media->llx, media->lly);
+	    }
+	  else
+	    {
+	      if (nup_landscape)
+		{
+		  output ("90 rotate\n");
+		  output ("%d %d translate\n", media->lly, -media->llx);
+		}
+	      else
+		output ("%d %d translate\n", media->llx, media->ury);
+	    }
+
+	  /*
+	   * Scale us down a bit so the page can be printed in its
+	   * bouding box.
+	   */
+	  output ("%f dup scale\n",
+		  (float) (media->urx - media->llx) / media->w);
+	}
+    }
 
   /* Page start comment. */
   switch (page_label)
     {
     case LABEL_SHORT:
-      output ("%%%%Page: (%d) %d\n", current_pagenum, total_pages);
+      output ("%sPage: (%d) %d\n", cstr, current_pagenum, total_pages);
       break;
 
     case LABEL_LONG:
-      output ("%%%%Page: (%s:%3d) %d\n", ftail, current_pagenum, total_pages);
+      output ("%sPage: (%s:%3d) %d\n", cstr, ftail, current_pagenum,
+	      total_pages);
       break;
     }
 
@@ -1534,14 +1755,34 @@ dump_ps_page_header (char *fname)
    * Page Setup.
    */
 
-  output ("%%%%BeginPageSetup\n");
+  output ("%sBeginPageSetup\n", cstr);
 
   output ("_S\n");
 
+  if (nup > 1)
+    {
+      output ("%% N-up sub-page %d/%d\n", nup_subpage + 1, nup);
+      if (landscape)
+	output ("%d %d translate\n",
+		(nup_subpage / nup_rows) * nup_width,
+		(nup_subpage % nup_rows) * nup_height);
+      else
+	output ("%d %d translate\n",
+		(nup_subpage % nup_columns) * nup_width,
+		- ((nup_subpage / nup_columns + 1) * nup_height));
+      output ("%g dup scale\n", nup_scale);
+    }
+
+#if PRINT_BOUNDING_BOXES
+	  output ("%d %d moveto %d %d lineto %d %d lineto %d %d lineto closepath stroke\n",
+		  media->llx, media->lly, media->llx, media->ury,
+		  media->urx, media->ury, media->urx, media->lly);
+#endif
+
   if (landscape)
     {
       output ("90 rotate\n");
-      output ("%d %d translate\n", media->lly, media->llx - media->w);
+      output ("%d %d translate\n", media->lly, -media->urx);
     }
   else
     output ("%d %d translate\n", media->llx, media->lly);
@@ -1552,10 +1793,11 @@ dump_ps_page_header (char *fname)
   output ("/ftail (%s) def\n", ftail);
 
   /* Do we have a pending ^@font{} font? */
-  if (user_font)
+  if (user_fontp)
     {
       output ("/%s /F-gs-user-font MF\n", Fname);
-      output ("/F-gs-user-font findfont %g scalefont setfont\n", Fpt);
+      output ("/F-gs-user-font findfont [%g 0 0 %g 0 0] makefont setfont\n",
+	      Fpt.w, Fpt.h);
     }
 
   /* Dump user defined strings. */
@@ -1566,7 +1808,7 @@ dump_ps_page_header (char *fname)
 	   got;
 	   got = strhash_get_next (user_strings, &cp, &i, (void **) &cp2))
 	{
-	  cp2 = format_user_string (cp2);
+	  cp2 = format_user_string ("%Format", cp2);
 	  output ("/%s (%s) def\n", cp, cp2);
 	  xfree (cp2);
 	}
@@ -1612,14 +1854,62 @@ dump_ps_page_header (char *fname)
       output ("do_header\n");
       break;
     }
+
+  /* Do we have a pending ^@color{} color? */
+  if (user_colorp)
+    output ("%g %g %g setrgbcolor\n", user_color.r, user_color.g,
+	    user_color.b);
 }
 
 
 static void
 dump_ps_page_trailer ()
 {
+  unsigned int nup_subpage = (total_pages - 1) % nup;
+
   output ("_R\n");
-  output ("S\n");
+
+  if (nup > 1)
+    {
+      if (nup_subpage + 1 == nup)
+	{
+	  /* Real end of page. */
+	  output ("_R\n");
+	  output ("S\n");
+	}
+    }
+  else
+    output ("S\n");
+}
+
+
+static void
+dump_empty_page ()
+{
+  if (nup > 1)
+    {
+      unsigned int nup_subpage = (total_pages - 1) % nup;
+
+      if (nup_subpage == 0)
+	{
+	  /* Real start of the page, must do it by the harder way. */
+	  dump_ps_page_header (fname);
+	}
+      else
+	output ("%%Page: (-) %d\n", total_pages);
+
+      if (nup_subpage + 1 == nup)
+	{
+	  /* This is the last page on this sheet, dump us. */
+	  output ("_R\n");
+	  output ("S\n");
+	}
+    }
+  else
+    {
+      output ("%%%%Page: (-) %d\n", total_pages);
+      output ("S\n");
+    }
 }
 
 
@@ -1633,7 +1923,7 @@ recognize_eps_file (Token *token)
   int valid_epsf;
   float llx, lly, urx, ury;
 
-  message (1, "^@epsf=\"%s\"\n", token->u.epsf.filename);
+  message (2, "^@epsf=\"%s\"\n", token->u.epsf.filename);
 
   i = strlen (token->u.epsf.filename);
   if (i > 0 && token->u.epsf.filename[i - 1] == '|')
