@@ -30,12 +30,15 @@
  * Static variables.
  */
 
-/* 7bit ASCII scand encodings (additions to 7bit ASCII enc). */
+/*
+ * 7bit ASCII fi(nland), se (sweden) scand encodings (additions to 7bit ASCII
+ * enc).
+ */
 static struct
 {
   int code;
   char *name;
-} enc_7bit_ascii_scands[] =
+} enc_7bit_ascii_fise[] =
 {
   {'{',		"adieresis"},
   {'|',		"odieresis"},
@@ -46,12 +49,31 @@ static struct
   {0, NULL},
 };
 
+/*
+ * 7bit ASCII dk (denmark), no(rway) scand encodings (additions to 7bit ASCII
+ * enc).
+ */
+static struct
+{
+  int code;
+  char *name;
+} enc_7bit_ascii_dkno[] =
+{
+  {'{',		"ae"},
+  {'|',		"oslash"},
+  {'}',		"aring"},
+  {'[',		"AE"},
+  {'\\',	"Oslash"},
+  {']',		"Aring"},
+  {0, NULL},
+};
+
 
 /*
  * Prototypes for static functions.
  */
 
-static void cfg_fatal __P ((char *file, int line, char *fmt, ...));
+static void cfg_fatal ___P ((char *file, int line, char *fmt, ...));
 
 
 /*
@@ -327,6 +349,12 @@ read_config (char *path, char *file)
 
 	  add_media (name, w, h, llx, lly, urx, ury);
 	}
+      else if (MATCH (token, "NoJobHeaderSwitch:"))
+	{
+	  token2 = GET_LINE_TOKEN (NULL);
+	  CHECK_TOKEN ();
+	  strcpy (no_job_header_switch, token2);
+	}
       else if (MATCH (token, "NonPrintableFormat:"))
 	{
 	  token2 = GET_TOKEN (NULL);
@@ -334,6 +362,12 @@ read_config (char *path, char *file)
 	  strcpy (npf_name_buf, token2);
 	  npf_name = npf_name_buf;
 	}
+      else if (MATCH (token, "OutputFirstLine:"))
+	{
+	  token2 = GET_LINE_TOKEN (NULL);
+	  CHECK_TOKEN ();
+	  strcpy (output_first_line, token2);
+	}
       else if (MATCH (token, "PageLabelFormat:"))
 	{
 	  token2 = GET_TOKEN (NULL);
@@ -372,12 +406,36 @@ read_config (char *path, char *file)
 	  CHECK_TOKEN ();
 	  strcpy (spooler_command, token2);
 	}
+      else if (MATCH (token, "StatesColorModel:"))
+	{
+	  token2 = GET_TOKEN (NULL);
+	  CHECK_TOKEN ();
+	  strcpy (states_color_model, token2);
+	}
+      else if (MATCH (token, "StatesConfigFile:"))
+	{
+	  token2 = GET_LINE_TOKEN (NULL);
+	  CHECK_TOKEN ();
+	  strcpy (states_config_file, token2);
+	}
+      else if (MATCH (token, "StatesHighlightLevel:"))
+	{
+	  token2 = GET_TOKEN (NULL);
+	  CHECK_TOKEN ();
+	  strcpy (states_highlight_level, token2);
+	}
       else if (MATCH (token, "StatusDict:"))
 	{
 	  token2 = GET_TOKEN (NULL);
 	  CHECK_TOKEN ();
 	  parse_key_value_pair (statusdict, token2);
 	}
+      else if (MATCH (token, "TOCFormat:"))
+	{
+	  token2 = GET_LINE_TOKEN (NULL);
+	  CHECK_TOKEN ();
+	  toc_fmt_string = xstrdup (token2);
+	}
       else if (MATCH (token, "Underlay:"))
 	{
 	  token2 = GET_LINE_TOKEN (NULL);
@@ -676,10 +734,10 @@ paste_file (char *name, char *suffix)
 
 
 int
-parse_font_spec (char *spec, char **name_return, double *size_return)
+parse_font_spec (char *spec, char **name_return, FontPoint *size_return)
 {
   int i;
-  char *cp;
+  char *cp, *cp2;
 
   /* The `name@ptsize' format? */
   cp = strchr (spec, '@');
@@ -689,8 +747,18 @@ parse_font_spec (char *spec, char **name_return, double *size_return)
       if (cp[1] == '\0')
 	/* No ptsize after '@'. */
 	return 0;
+      cp++;
 
-      *size_return = atof (cp + 1);
+      /* Check pt/pt format. */
+      cp2 = strchr (cp, '/');
+      if (cp2)
+	{
+	  *cp2++ = '\0';
+	  size_return->w = atof (cp);
+	  size_return->h = atof (cp2);
+	}
+      else
+	size_return->w = size_return->h = atof (cp);
     }
   else
     {
@@ -701,15 +769,41 @@ parse_font_spec (char *spec, char **name_return, double *size_return)
 
       for (i--; i >= 0 && ISNUMBERDIGIT (spec[i]); i--)
 	;
-      i++;
-      *size_return = atof (spec + i);
+      if (i < 0)
+	return 0;
+
+      /* Check pt/pt format. */
+      if (spec[i] == '/')
+	{
+	  size_return->h = atof (spec + i + 1);
+
+	  for (i--; i >= 0 && ISNUMBERDIGIT (spec[i]); i--)
+	    ;
+	  if (i < 0)
+	    return 0;
+	  i++;
+	  size_return->w = atof (spec + i);
+	}
+      else
+	{
+	  i++;
+	  size_return->w = size_return->h = atof (spec + i);
+	}
     }
   
   *name_return = (char *) xcalloc (1, i + 1);
   strncpy (*name_return, spec, i);
 
-  message (2, "parse_font_spec(): name=%.*s, size=%g\n", i, *name_return,
-	   *size_return);
+  message (2, "parse_font_spec(): name=%.*s, size=%g/%g\n", i, *name_return,
+	   size_return->w, size_return->h);
+
+  if (size_return->w < 0.0 && size_return->h < 0.0)
+    message (0, _("%s: warning: font size is negative\n"), program);
+  else if (size_return->w < 0.0)
+    message (0, _("%s: warning: font width is negative\n"), program);
+  else if (size_return->h < 0.0)
+    message (0, _("%s: warning: font height is negative\n"), program);
+    
   return 1;
 }
 
@@ -717,119 +811,164 @@ parse_font_spec (char *spec, char **name_return, double *size_return)
 void
 read_font_info (void)
 {
+  CachedFontInfo *font_info;
   AFMFont font;
-  int cached = 1;
+  int font_info_cached = 1;
+  int font_cached = 1;
   int i;
   unsigned int enc_flags = 0;
+  char fkey[256];
 
-  message (1, _("reading AFM info for font \"%s\"\n"), Fname);
+  message (2, _("reading AFM info for font \"%s\"\n"), Fname);
 
   if (accept_composites)
     enc_flags = AFM_ENCODE_ACCEPT_COMPOSITES;
 
   /* Open font */
-  if (!strhash_get (afm_cache, Fname, strlen (Fname), (void **) &font))
+  sprintf (fkey, "%s@%f", Fname, Fpt.w);
+  if (!strhash_get (afm_info_cache, fkey, strlen (fkey), (void **) &font_info))
     {
       AFMError error;
       char buf[256];
 
-      /* Couldn't find it from our cache, open it from disk. */
-      error = afm_open_font (afm, AFM_I_COMPOSITES, Fname, &font);
-      if (error != AFM_SUCCESS)
+      /* Couldn't find it from our cache, open open AFM file. */
+      if (!strhash_get (afm_cache, Fname, strlen (Fname), (void **) &font))
 	{
-#define COUR "Courier"
-	  /*
-	   * Do not report failures for "Courier*" fonts because 
-	   * AFM library's default font will fix them.
-	   */
-	  if (strncmp (Fname, COUR, strlen (COUR)) != 0)
-	    message (0, _("couldn't open AFM file for font \"%s\", using default\n"),
-		     Fname);
-	  error = afm_open_default_font (afm, &font);
+	  /* AFM file was not cached, open it from disk. */
+	  error = afm_open_font (afm, AFM_I_COMPOSITES, Fname, &font);
 	  if (error != AFM_SUCCESS)
 	    {
-	      afm_error_to_string (error, buf);
-	      fatal (_("couldn't open AFM file for the default font: %s"),
-		     buf);
+#define COUR "Courier"
+	      /*
+	       * Do not report failures for "Courier*" fonts because 
+	       * AFM library's default font will fix them.
+	       */
+	      if (strncmp (Fname, COUR, strlen (COUR)) != 0)
+		message (0, _("couldn't open AFM file for font \"%s\", using default\n"),
+			 Fname);
+	      error = afm_open_default_font (afm, &font);
+	      if (error != AFM_SUCCESS)
+		{
+		  afm_error_to_string (error, buf);
+		  fatal (_("couldn't open AFM file for the default font: %s"),
+			 buf);
+		}
 	    }
-	}
 
-      /* Apply encoding. */
-      switch (encoding)
+	  /* Apply encoding. */
+	  switch (encoding)
+	    {
+	    case ENC_LATIN1:
+	      (void) afm_font_encoding (font, AFM_ENCODING_LATIN1, enc_flags);
+	      break;
+
+	    case ENC_LATIN2:
+	      (void) afm_font_encoding (font, AFM_ENCODING_LATIN2, enc_flags);
+	      break;
+
+	    case ENC_LATIN3:
+	      (void) afm_font_encoding (font, AFM_ENCODING_LATIN3, enc_flags);
+	      break;
+
+	    case ENC_LATIN5:
+	      (void) afm_font_encoding (font, AFM_ENCODING_LATIN5, enc_flags);
+	      break;
+
+	    case ENC_ASCII:
+	      (void) afm_font_encoding (font, AFM_ENCODING_ASCII, enc_flags);
+	      break;
+
+	    case ENC_ASCII_FISE:
+	      /* First apply standard 7bit ASCII encoding. */
+	      (void) afm_font_encoding (font, AFM_ENCODING_ASCII, enc_flags);
+
+	      /* Then add those scand characters. */
+	      for (i = 0; enc_7bit_ascii_fise[i].name; i++)
+		(void) afm_font_encode (font, enc_7bit_ascii_fise[i].code,
+					enc_7bit_ascii_fise[i].name,
+					enc_flags);
+	      break;
+
+	    case ENC_ASCII_DKNO:
+	      /* First apply standard 7bit ASCII encoding. */
+	      (void) afm_font_encoding (font, AFM_ENCODING_ASCII, enc_flags);
+
+	      /* Then add those scand characters. */
+	      for (i = 0; enc_7bit_ascii_dkno[i].name; i++)
+		(void) afm_font_encode (font, enc_7bit_ascii_dkno[i].code,
+					enc_7bit_ascii_dkno[i].name,
+					enc_flags);
+	      break;
+
+	    case ENC_IBMPC:
+	      (void) afm_font_encoding (font, AFM_ENCODING_IBMPC, enc_flags);
+	      break;
+
+	    case ENC_MAC:
+	      (void) afm_font_encoding (font, AFM_ENCODING_MAC, enc_flags);
+	      break; 
+
+	    case ENC_VMS:
+	      (void) afm_font_encoding (font, AFM_ENCODING_VMS, enc_flags);
+	      break; 
+
+	    case ENC_HP8:
+	      (void) afm_font_encoding (font, AFM_ENCODING_HP8, enc_flags);
+	      break;
+
+	    case ENC_KOI8:
+	      (void) afm_font_encoding (font, AFM_ENCODING_KOI8, enc_flags);
+	      break;
+
+	    case ENC_PS:
+	      /* Let's use font's default encoding -- nothing here. */
+	      break;
+	    }
+
+	  /* Put it to the AFM cache. */
+	  if (!strhash_put (afm_cache, Fname, strlen (Fname), font, NULL))
+	    font_cached = 0;
+	}
+      
+      font_info = (CachedFontInfo *) xcalloc (1, sizeof (*font_info));
+      /* Read character widths and types. */
+      for (i = 0; i < 256; i++)
 	{
-	case ENC_LATIN1:
-	  (void) afm_font_encoding (font, AFM_ENCODING_LATIN1, enc_flags);
-	break;
+	  AFMNumber w0x, w0y;
 
-	case ENC_LATIN2:
-	  (void) afm_font_encoding (font, AFM_ENCODING_LATIN2, enc_flags);
-	break;
-
-	case ENC_LATIN3:
-	  (void) afm_font_encoding (font, AFM_ENCODING_LATIN3, enc_flags);
-	break;
-
-	case ENC_ASCII:
-	  (void) afm_font_encoding (font, AFM_ENCODING_ASCII, enc_flags);
-	break;
-
-	case ENC_ASCII_SCANDS:
-	  /* First apply standard 7bit ASCII encoding. */
-	  (void) afm_font_encoding (font, AFM_ENCODING_ASCII, enc_flags);
-
-	/* Then add those scand characters. */
-	for (i = 0; enc_7bit_ascii_scands[i].name; i++)
-	  (void) afm_font_encode (font, enc_7bit_ascii_scands[i].code,
-				  enc_7bit_ascii_scands[i].name, enc_flags);
-	break;
-
-	case ENC_IBMPC:
-	  (void) afm_font_encoding (font, AFM_ENCODING_IBMPC, enc_flags);
-	break;
-
-	case ENC_MAC:
-	  (void) afm_font_encoding (font, AFM_ENCODING_MAC, enc_flags);
-	break; 
-
-	case ENC_VMS:
-	  (void) afm_font_encoding (font, AFM_ENCODING_VMS, enc_flags);
-	break; 
-
-	case ENC_HP8:
-	  (void) afm_font_encoding (font, AFM_ENCODING_HP8, enc_flags);
-	break;
-
-	case ENC_PS:
-	  /* Let's use font's default encoding -- nothing here. */
-	  break;
+	  (void) afm_font_charwidth (font, Fpt.w, (unsigned int) i,
+				     &w0x, &w0y);
+	  font_info->font_widths[i] = w0x;
+	      
+	  if (font->encoding[i] == AFM_ENC_NONE)
+	    font_info->font_ctype[i] = ' ';
+	  else if (font->encoding[i] == AFM_ENC_NON_EXISTENT)
+	    font_info->font_ctype[i] = '.';
+	  else
+	    font_info->font_ctype[i] = '*';
 	}
 
-      /* Store font information to the AFM cache. */
-      if (!strhash_put (afm_cache, Fname, strlen (Fname), font, NULL))
-	cached = 0;
+      font_info->font_is_fixed
+	= font->writing_direction_metrics[0].IsFixedPitch;
+      font_info->font_bbox_lly = font->global_info.FontBBox_lly;
+
+      if (!font_cached)
+	(void) afm_close_font (font);
+
+      /* Store font information to the AFM information cache. */
+      if (!strhash_put (afm_info_cache, fkey, strlen (fkey), font_info, NULL))
+	font_info_cached = 0;
     }
 
-  /* Read character widths and types. */
-  for (i = 0; i < 256; i++)
-    {
-      AFMNumber w0x, w0y;
+  /* Select character widths and types. */
+  memcpy (font_widths, font_info->font_widths, 256 * sizeof (double));
+  memcpy (font_ctype, font_info->font_ctype, 256);
 
-      (void) afm_font_charwidth (font, Fpt, (unsigned int) i, &w0x, &w0y);
-      font_widths[i] = w0x;
-	      
-      if (font->encoding[i] == AFM_ENC_NONE)
-	font_ctype[i] = ' ';
-      else if (font->encoding[i] == AFM_ENC_NON_EXISTENT)
-	font_ctype[i] = '.';
-      else
-	font_ctype[i] = '*';
-    }
+  font_is_fixed = font_info->font_is_fixed;
+  font_bbox_lly = font_info->font_bbox_lly;
 
-  font_is_fixed = font->writing_direction_metrics[0].IsFixedPitch;
-  font_bbox_lly = font->global_info.FontBBox_lly;
-
-  if (!cached)
-    (void) afm_close_font (font);
+  if (!font_info_cached)
+    xfree (font_info);
 }
 
 
@@ -1048,20 +1187,41 @@ escape_string (char *string)
 
 #define APPEND_CH(ch)				\
   do {						\
-    NEED_NBYTES (1);				\
-    rbuf[rbufpos++] = (ch);			\
+    int a;					\
+    NEED_NBYTES (width);			\
+    if (width && justification < 0)		\
+      rbuf[rbufpos++] = (ch);			\
+    for (a = 0; a < width - 1; a++)		\
+      rbuf[rbufpos++] = ' ';			\
+    if (!width || justification > 0)		\
+      rbuf[rbufpos++] = (ch);			\
   } while (0)
 
 #define APPEND_STR(str)				\
   do {						\
     int len = strlen ((str));			\
-    NEED_NBYTES (len);				\
+    int nspace;					\
+						\
+    if (len > width)				\
+      nspace = 0;				\
+    else					\
+      nspace = width - len;			\
+						\
+    NEED_NBYTES (nspace + len);			\
+    if (width && justification > 0)		\
+      for (; nspace; nspace--)			\
+	rbuf[rbufpos++] = ' ';			\
+						\
     memcpy (rbuf + rbufpos, str, len);		\
     rbufpos += len;				\
+						\
+    if (width && justification < 0)		\
+      for (; nspace; nspace--)			\
+	rbuf[rbufpos++] = ' ';			\
   } while (0)
 
 char *
-format_user_string (char *str)
+format_user_string (char *context_name, char *str)
 {
   char *cp;
   char *rbuf = NULL;
@@ -1071,256 +1231,312 @@ format_user_string (char *str)
   int j;
   char buf[512];
   char buf2[512];
+  int width = 0;
+  int justification = 1;
 
   /* Format string. */
   for (i = 0; str[i] != '\0'; i++)
     {
-      switch (str[i])
+      int type;
+
+      type = str[i];
+
+      if (type == '%' || type == '$')
 	{
-	case '%':		/* general state related `%' escapes  */
 	  i++;
-	  switch (str[i])
+	  width = 0;
+	  justification = 1;
+
+	  /* Get optional width and justification. */
+	  if (str[i] == '-')
 	    {
-	    case '%':		/* `%%' character `%' */
-	      APPEND_CH ('%');
-	      break;
+	      i++;
+	      justification = -1;
+	    }
+	  while (isdigit (str[i]))
+	    width = width * 10 + str[i++] - '0';
 
-	    case 'c':		/* `%c' trailing component of pwd. */
-	      getcwd (buf, sizeof (buf));
-	      cp = strrchr (buf, '/');
-	      if (cp)
-		cp++;
-	      else
-		cp = buf;
-	      APPEND_STR (cp);
-	      break;
-
-	    case 'C':		/* `%C' runtime in `hh:mm:ss' format */
-	      sprintf (buf, "%02d:%02d:%02d", run_tm.tm_hour, run_tm.tm_min,
-		       run_tm.tm_sec);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'd':		/* `%d' current working directory */
-	      getcwd (buf, sizeof (buf));
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'D':
-	      if (str[i + 1] == '{')
+	  /* Handle escapes. */
+	  if (type == '%')
+	    {
+	      /* General state related %-escapes. */
+	      switch (str[i])
 		{
-		  /* `%D{}' format run date with strftime() */
-		  for (j = 0, i += 2;
-		       j < sizeof (buf2) && str[i] && str[i] != '}';
-		       i++, j++)
-		    buf2[j] = str[i];
-		  if (str[i] != '}')
-		    fatal (_("%%Format: too long format for %%D{} escape"));
+		case '%':	/* `%%' character `%' */
+		  APPEND_CH ('%');
+		  break;
 
-		  buf2[j] = '\0';
-		  strftime (buf, sizeof (buf), buf2, &run_tm);
-		}
-	      else
-		{
-		  /* `%D' run date in `yy-mm-dd' format */
-		  sprintf (buf, "%02d-%02d-%02d", run_tm.tm_year,
+		case 'c':	/* `%c' trailing component of pwd. */
+		  getcwd (buf, sizeof (buf));
+		  cp = strrchr (buf, '/');
+		  if (cp)
+		    cp++;
+		  else
+		    cp = buf;
+		  APPEND_STR (cp);
+		  break;
+
+		case 'C':	/* `%C' runtime in `hh:mm:ss' format */
+		  sprintf (buf, "%02d:%02d:%02d", run_tm.tm_hour,
+			   run_tm.tm_min, run_tm.tm_sec);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'd':	/* `%d' current working directory */
+		  getcwd (buf, sizeof (buf));
+		  APPEND_STR (buf);
+		  break;
+
+		case 'D':
+		  if (str[i + 1] == '{')
+		    {
+		      /* `%D{}' format run date with strftime() */
+		      for (j = 0, i += 2;
+			   j < sizeof (buf2) && str[i] && str[i] != '}';
+			   i++, j++)
+			buf2[j] = str[i];
+		      if (str[i] != '}')
+			fatal (_("%s: too long format for %%D{} escape"),
+			       context_name);
+
+		      buf2[j] = '\0';
+		      strftime (buf, sizeof (buf), buf2, &run_tm);
+		    }
+		  else
+		    {
+		      /* `%D' run date in `yy-mm-dd' format */
+		      sprintf (buf, "%02d-%02d-%02d", run_tm.tm_year,
+			       run_tm.tm_mon + 1, run_tm.tm_mday);
+		    }
+		  APPEND_STR (buf);
+		  break;
+
+		case 'E':	/* `%E' run date in `yy/mm/dd' format */
+		  sprintf (buf, "%02d/%02d/%02d", run_tm.tm_year,
 			   run_tm.tm_mon + 1, run_tm.tm_mday);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'F':	/* `%F' run date in `dd.mm.yyyy' format */
+		  sprintf (buf, "%d.%d.%d",
+			   run_tm.tm_mday,
+			   run_tm.tm_mon + 1,
+			   run_tm.tm_year+1900);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'H':	/* `%H' document title */
+		  APPEND_STR (title);
+		  break;
+
+		case 'm':	/* `%m' the hostname up to the first `.' */
+		  (void) gethostname (buf, sizeof (buf));
+		  cp = strchr (buf, '.');
+		  if (cp)
+		    *cp = '\0';
+		  APPEND_STR (buf);
+		  break;
+
+		case 'M':	/* `%M' the full hostname */
+		  (void) gethostname (buf, sizeof (buf));
+		  APPEND_STR (buf);
+		  break;
+
+		case 'n':	/* `%n' username */
+		  APPEND_STR (passwd->pw_name);
+		  break;
+
+		case 'N':	/* `%N' pw_gecos up to the first `,' char */
+		  strcpy (buf, passwd->pw_gecos);
+		  cp = strchr (buf, ',');
+		  if (cp)
+		    *cp = '\0';
+		  APPEND_STR (buf);
+		  break;
+
+		case 't':	/* `%t' runtime in 12-hour am/pm format */
+		  sprintf (buf, "%d:%d%s",
+			   run_tm.tm_hour > 12
+			   ? run_tm.tm_hour - 12 : run_tm.tm_hour,
+			   run_tm.tm_min,
+			   run_tm.tm_hour > 12 ? "pm" : "am");
+		  APPEND_STR (buf);
+		  break;
+
+		case 'T':	/* `%T' runtime in 24-hour format */
+		  sprintf (buf, "%d:%d", run_tm.tm_hour, run_tm.tm_min);
+		  APPEND_STR (buf);
+		  break;
+
+		case '*':	/* `%*' runtime in 24-hour format with secs */
+		  sprintf (buf, "%d:%d:%d", run_tm.tm_hour, run_tm.tm_min,
+			   run_tm.tm_sec);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'W':	/* `%W' run date in `mm/dd/yy' format */
+		  sprintf (buf, "%02d/%02d/%02d", run_tm.tm_mon + 1,
+			   run_tm.tm_mday, run_tm.tm_year);
+		  APPEND_STR (buf);
+		  break;
+
+		default:
+		  fatal (_("%s: unknown `%%' escape `%c' (%d)"), context_name,
+			 str[i], str[i]);
+		  break;
 		}
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'E':		/* `%E' run date in `yy/mm/dd' format */
-	      sprintf (buf, "%02d/%02d/%02d", run_tm.tm_year,
-		       run_tm.tm_mon + 1, run_tm.tm_mday);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'F':		/* `%F' run date in `dd.mm.yyyy' format */
-	      sprintf (buf, "%d.%d.%d",
-		       run_tm.tm_mday,
-		       run_tm.tm_mon + 1,
-		       run_tm.tm_year+1900);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'H':		/* `%H' document title */
-	      APPEND_STR (title);
-	      break;
-
-	    case 'm':		/* `%m' the hostname up to the first `.' */
-	      (void) gethostname (buf, sizeof (buf));
-	      cp = strchr (buf, '.');
-	      if (cp)
-		*cp = '\0';
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'M':		/* `%M' the full hostname */
-	      (void) gethostname (buf, sizeof (buf));
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'n':		/* `%n' username */
-	      APPEND_STR (passwd->pw_name);
-	      break;
-
-	    case 'N':		/* `%N' pw_gecos up to the first `,' char */
-	      strcpy (buf, passwd->pw_gecos);
-	      cp = strchr (buf, ',');
-	      if (cp)
-		*cp = '\0';
-	      APPEND_STR (buf);
-	      break;
-
-	    case 't':		/* `%t' runtime in 12-hour am/pm format */
-	      sprintf (buf, "%d:%d%s",
-		       run_tm.tm_hour > 12
-		       ? run_tm.tm_hour - 12 : run_tm.tm_hour,
-		       run_tm.tm_min,
-		       run_tm.tm_hour > 12 ? "pm" : "am");
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'T':		/* `%T' runtime in 24-hour format */
-	      sprintf (buf, "%d:%d", run_tm.tm_hour, run_tm.tm_min);
-	      APPEND_STR (buf);
-	      break;
-
-	    case '*':		/* `%*' runtime in 24-hour format with secs */
-	      sprintf (buf, "%d:%d:%d", run_tm.tm_hour, run_tm.tm_min,
-		       run_tm.tm_sec);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'W':		/* `%W' run date in `mm/dd/yy' format */
-	      sprintf (buf, "%02d/%02d/%02d", run_tm.tm_mon + 1,
-		       run_tm.tm_mday, run_tm.tm_year);
-	      APPEND_STR (buf);
-	      break;
-
-	    default:
-	      fatal (_("%%Format: unknown `%%' escape `%c' (%d)"),
-		     str[i], str[i]);
-	      break;
 	    }
-	  break;
-
-	case '$':		/* input file related `$' escapes */
-	  i++;
-	  switch (str[i])
+	  else
 	    {
-	    case '$':		/* `$$' character `$' */
-	      APPEND_CH ('$');
-	      break;
-
-	    case '%':		/* `$%' current page number */
-	      sprintf (buf, "%d", current_pagenum);
-	      APPEND_STR (buf);
-	      break;
-
-	    case '=':		/* `$=' number of pages in this file */
-	      APPEND_CH ('\001');
-	      break;
-
-	    case '(':		/* $(ENVVAR)  */
-	      for (j = 0, i++;
-		   str[i] && str[i] != ')' && j < sizeof (buf) - 1;
-		   i++)
-		buf[j++] = str[i];
-
-	      if (str[i] == '\0')
-		fatal (_("%%Format: no closing ')' for $() escape"));
-	      if (str[i] != ')')
-		fatal (_("%%Format: too long variable name for $() escape"));
-
-	      buf[j] = '\0';
-
-	      cp = getenv (buf);
-	      if (cp == NULL)
-		cp = "";
-	      APPEND_STR (cp);
-	      break;
-
-	    case 'C':		/* `$C' modtime in `hh:mm:ss' format */
-	      sprintf (buf, "%02d:%02d:%02d", mod_tm.tm_hour,
-		       mod_tm.tm_min, mod_tm.tm_sec);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'D':
-	      if (str[i + 1] == '{')
+	      /* Input file related $-escapes. */
+	      switch (str[i])
 		{
-		  /* `$D{}' format modification date with strftime() */
-		  for (j = 0, i += 2;
-		       j < sizeof (buf2) && str[i] && str[i] != '}';
-		       i++, j++)
-		    buf2[j] = str[i];
-		  if (str[i] != '}')
-		    fatal (_("%%Format: too long format for $D{} escape"));
+		case '$':	/* `$$' character `$' */
+		  APPEND_CH ('$');
+		  break;
 
-		  buf2[j] = '\0';
-		  strftime (buf, sizeof (buf), buf2, &mod_tm);
-		}
-	      else
-		{
-		  /* `$D' mod date in `yy-mm-dd' format */
-		  sprintf (buf, "%02d-%02d-%02d", mod_tm.tm_year,
+		case '%':	/* `$%' current page number */
+		  sprintf (buf, "%d", current_pagenum);
+		  APPEND_STR (buf);
+		  break;
+
+		case '=':	/* `$=' number of pages in this file */
+		  APPEND_CH ('\001');
+		  break;
+
+		case '(':	/* $(ENVVAR)  */
+		  for (j = 0, i++;
+		       str[i] && str[i] != ')' && j < sizeof (buf) - 1;
+		       i++)
+		    buf[j++] = str[i];
+
+		  if (str[i] == '\0')
+		    fatal (_("%s: no closing ')' for $() escape"),
+			   context_name);
+		  if (str[i] != ')')
+		    fatal (_("%s: too long variable name for $() escape"),
+			   context_name);
+
+		  buf[j] = '\0';
+
+		  cp = getenv (buf);
+		  if (cp == NULL)
+		    cp = "";
+		  APPEND_STR (cp);
+		  break;
+
+		case 'C':	/* `$C' modtime in `hh:mm:ss' format */
+		  sprintf (buf, "%02d:%02d:%02d", mod_tm.tm_hour,
+			   mod_tm.tm_min, mod_tm.tm_sec);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'D':
+		  if (str[i + 1] == '{')
+		    {
+		      /* `$D{}' format modification date with strftime() */
+		      for (j = 0, i += 2;
+			   j < sizeof (buf2) && str[i] && str[i] != '}';
+			   i++, j++)
+			buf2[j] = str[i];
+		      if (str[i] != '}')
+			fatal (_("%s: too long format for $D{} escape"),
+			       context_name);
+
+		      buf2[j] = '\0';
+		      strftime (buf, sizeof (buf), buf2, &mod_tm);
+		    }
+		  else
+		    {
+		      /* `$D' mod date in `yy-mm-dd' format */
+		      sprintf (buf, "%02d-%02d-%02d", mod_tm.tm_year,
+			       mod_tm.tm_mon + 1, mod_tm.tm_mday);
+		    }
+		  APPEND_STR (buf);
+		  break;
+
+		case 'E':	/* `$E' mod date in `yy/mm/dd' format */
+		  sprintf (buf, "%02d/%02d/%02d", mod_tm.tm_year,
 			   mod_tm.tm_mon + 1, mod_tm.tm_mday);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'F':	/* `$F' run date in `dd.mm.yyyy' format */
+		  sprintf (buf, "%d.%d.%d",
+			   mod_tm.tm_mday,
+			   mod_tm.tm_mon + 1,
+			   mod_tm.tm_year+1900);
+		  APPEND_STR (buf);
+		  break;
+
+		case 't':	/* `$t' runtime in 12-hour am/pm format */
+		  sprintf (buf, "%d:%d%s",
+			   mod_tm.tm_hour > 12
+			   ? mod_tm.tm_hour - 12 : mod_tm.tm_hour,
+			   mod_tm.tm_min,
+			   mod_tm.tm_hour > 12 ? "pm" : "am");
+		  APPEND_STR (buf);
+		  break;
+
+		case 'T':	/* `$T' runtime in 24-hour format */
+		  sprintf (buf, "%d:%d", mod_tm.tm_hour, mod_tm.tm_min);
+		  APPEND_STR (buf);
+		  break;
+
+		case '*':	/* `$*' runtime in 24-hour format with secs */
+		  sprintf (buf, "%d:%d:%d", mod_tm.tm_hour, mod_tm.tm_min,
+			   mod_tm.tm_sec);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'v':	/* `$v': input file number */
+		  sprintf (buf, "%d", input_filenum);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'V':	/* `$V': input file number in --toc format */
+		  if (toc)
+		    {
+		      sprintf (buf, "%d-", input_filenum);
+		      APPEND_STR (buf);
+		    }
+		  break;
+
+		case 'W':	/* `$W' run date in `mm/dd/yy' format */
+		  sprintf (buf, "%02d/%02d/%02d", mod_tm.tm_mon + 1,
+			   mod_tm.tm_mday, mod_tm.tm_year);
+		  APPEND_STR (buf);
+		  break;
+
+		case 'N':	/* `$N' the full name of the printed file */
+		  APPEND_STR (fname);
+		  break;
+
+		case 'n':	/* `$n' input file name without directory */
+		  cp = strrchr (fname, '/');
+		  if (cp == NULL)
+		    cp = fname;
+		  APPEND_STR (cp);
+		  break;
+
+		case 'L':	/* `$L' number of lines in this file. */
+		  /* This is valid only for TOC-strings. */
+		  sprintf (buf, "%d", current_file_linenum - 1);
+		  APPEND_STR (buf);
+		  break;
+
+		default:
+		  fatal (_("%s: unknown `$' escape `%c' (%d)"), context_name,
+			 str[i], str[i]);
+		  break;
 		}
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'E':		/* `$E' mod date in `yy/mm/dd' format */
-	      sprintf (buf, "%02d/%02d/%02d", mod_tm.tm_year,
-		       mod_tm.tm_mon + 1, mod_tm.tm_mday);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'F':		/* `$F' run date in `dd.mm.yyyy' format */
-	      sprintf (buf, "%d.%d.%d",
-		       mod_tm.tm_mday,
-		       mod_tm.tm_mon + 1,
-		       mod_tm.tm_year+1900);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 't':		/* `$t' runtime in 12-hour am/pm format */
-	      sprintf (buf, "%d:%d%s",
-		       mod_tm.tm_hour > 12
-		       ? mod_tm.tm_hour - 12 : mod_tm.tm_hour,
-		       mod_tm.tm_min,
-		       mod_tm.tm_hour > 12 ? "pm" : "am");
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'T':		/* `$T' runtime in 24-hour format */
-	      sprintf (buf, "%d:%d", mod_tm.tm_hour, mod_tm.tm_min);
-	      APPEND_STR (buf);
-	      break;
-
-	    case '*':		/* `$*' runtime in 24-hour format with secs */
-	      sprintf (buf, "%d:%d:%d", mod_tm.tm_hour, mod_tm.tm_min,
-		       mod_tm.tm_sec);
-	      APPEND_STR (buf);
-	      break;
-
-	    case 'W':		/* `$W' run date in `mm/dd/yy' format */
-	      sprintf (buf, "%02d/%02d/%02d", mod_tm.tm_mon + 1,
-		       mod_tm.tm_mday, mod_tm.tm_year);
-	      APPEND_STR (buf);
-	      break;
-
-	    default:
-	      fatal (_("%%Format: unknown `$' escape `%c' (%d)"),
-		     str[i], str[i]);
-	      break;
 	    }
-	  break;
-
-	default:
-	  APPEND_CH (str[i]);
-	  break;
+	  /* Reset width so the else-arm goes ok at the next round. */
+	  width = 0;
+	  justification = 1;
 	}
+      else
+	APPEND_CH (str[i]);
     }
   APPEND_CH ('\0');
   
